from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.sim_ctrl.datatypes import DigitalSignal
from anasymod.structures.structure_config import StructureConfig
from anasymod.util import back2fwd
from anasymod.enums import FPGASimCtrl
from anasymod.templates.zynq_gpio import TemplZynqGPIO

def digsig(name, width=1, signed=False):
    # convenience function to return a digital signal
    return DigitalSignal(
        name=f'{name}',
        width=width,
        signed=signed,
        abspath=''
    )

class ModuleTop(JinjaTempl):
    """
    This is the generator for top.sv.
    """
    def __init__(self, target):
        super().__init__(trim_blocks=True, lstrip_blocks=True)
        scfg = target.str_cfg
        """ :type: StructureConfig """

        pcfg = target.prj_cfg
        """ :type: EmuConfig """

        self.result_path_raw = back2fwd(target.result_path_raw)

        #####################################################
        # Retrieve general configuration settings
        #####################################################

        # TODO: clean this up; there is no need to generate
        # the time manager infrastructure if it isn't going
        # to be instantiated.  Plus, it is a bit odd that
        # the emu_clk frequency ends up being twice what is
        # expected, when this option is set.
        self.no_time_manager = pcfg.cfg.no_time_manager

        #####################################################
        # Add plugin specific includes
        #####################################################

        self.plugin_includes = SVAPI()
        for plugin in target.plugins:
            for include_statement in plugin.include_statements:
                self.plugin_includes.writeln(f'{include_statement}')

        #####################################################
        # Create module interface
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name='top')
        module.add_inputs(scfg.clk_i)
        if ((target.cfg.fpga_sim_ctrl is not None) and
                (target.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ) and
                (not pcfg.board.is_ultrascale)):
            module.add_inouts(TemplZynqGPIO.EXT_IOS)
        module.generate_header()

        #####################################################
        # Manage clks
        #####################################################

        # Add clk in signals for simulation case
        self.clk_in_sim_sigs = SVAPI()

        for clk_i in scfg.clk_i:
            self.clk_in_sim_sigs.gen_signal(io_obj=clk_i)

        # Add dbg clk signals
        self.dbg_clk_sigs = SVAPI()
        self.dbg_clk_sigs.gen_signal(io_obj=scfg.dbg_clk)

        # Instantiation of clk_gen wrapper
        self.clk_gen_ifc = SVAPI()
        clk_gen = ModuleInst(api=self.clk_gen_ifc, name='clk_gen')
        clk_gen.add_inputs(scfg.clk_i, connections=scfg.clk_i)
        clk_gen.add_output(scfg.emu_clk_2x, connection=scfg.emu_clk_2x)
        clk_gen.add_output(scfg.dbg_clk, connection=scfg.dbg_clk)
        clk_gen.add_outputs(scfg.clk_independent, connections=scfg.clk_independent)
        clk_gen.generate_instantiation()

        # Absolute path assignments for derived clks
        self.derived_clk_assigns = SVAPI()
        for k, clk in enumerate(scfg.clk_derived):
            self.derived_clk_assigns.writeln(f'// derived clock: {clk.name}')
            if clk.abspath_emu_dt is not None:
                self.derived_clk_assigns.writeln(f'assign {clk.abspath_emu_dt} = emu_dt;')
            if clk.abspath_emu_clk is not None:
                self.derived_clk_assigns.writeln(f'assign {clk.abspath_emu_clk} = emu_clk;')
            if clk.abspath_emu_rst is not None:
                self.derived_clk_assigns.writeln(f'assign {clk.abspath_emu_rst} = emu_rst;')
            if clk.abspath_dt_req is not None:
                self.derived_clk_assigns.writeln(f'assign dt_req_{clk.name} = {clk.abspath_dt_req};')
            if clk.abspath_gated_clk is not None:
                self.derived_clk_assigns.writeln(f'assign clk_val_{clk.name} = {clk.abspath_gated_clk_req};')
                self.derived_clk_assigns.writeln(f'assign {clk.abspath_gated_clk} = clk_{clk.name};')

        self.num_dt_reqs = scfg.num_dt_reqs
        self.num_gated_clks = scfg.num_gated_clks

        #####################################################
        # Manage Ctrl Module
        #####################################################

        # provide information if default oscillator was used for template evaluation
        self.use_default_oscillator = scfg.use_default_oscillator

        ctrl_ios = scfg.analog_ctrl_inputs + scfg.analog_ctrl_outputs + scfg.digital_ctrl_inputs + \
                          scfg.digital_ctrl_outputs

        ## Instantiate all ctrl signals
        self.inst_itl_ctlsigs = SVAPI()
        for ctrl_io in ctrl_ios:
            self.inst_itl_ctlsigs.gen_signal(io_obj=ctrl_io)

        ## Instantiate ctrl module
        self.sim_ctrl_inst_ifc = SVAPI()
        sim_ctrl_inst = ModuleInst(api=self.sim_ctrl_inst_ifc, name='sim_ctrl_gen')
        sim_ctrl_inst.add_inputs(
            scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs,
            connections=scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs
        )
        sim_ctrl_inst.add_outputs(
            scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs,
            connections=scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        )

        ## Wire through Zynq connections if needed
        if ((target.cfg.fpga_sim_ctrl is not None) and
                (target.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ) and
                (not pcfg.board.is_ultrascale)):
            sim_ctrl_inst.add_inouts(TemplZynqGPIO.EXT_IOS,
                                     connections=TemplZynqGPIO.EXT_IOS)

        # add master clk to ctrl module
        emu_clk_sig = DigitalSignal(name='emu_clk', width=1, abspath=None)
        sim_ctrl_inst.add_input(emu_clk_sig, emu_clk_sig)
        sim_ctrl_inst.generate_instantiation()

        ## Assign custom ctrl signals via abs paths into design
        self.assign_custom_ctlsigs = SVAPI()
        for ctrl_input in scfg.digital_ctrl_inputs + scfg.analog_ctrl_inputs:
            if ctrl_input.abs_path is not None:
                self.assign_custom_ctlsigs.assign_to(io_obj=ctrl_input.abs_path, exp=ctrl_input.name)

        for ctrl_output in scfg.digital_ctrl_outputs + scfg.analog_ctrl_outputs:
            self.assign_custom_ctlsigs.assign_to(io_obj=ctrl_output.name, exp=ctrl_output.abs_path)

        ######################################################
        # Manage trace port Module
        ######################################################

        probes = scfg.digital_probes + scfg.analog_probes + [scfg.time_probe] + [scfg.dec_cmp]

        ## Instantiate all probe signals
        self.inst_probesigs = SVAPI()
        for probe in probes:
            self.inst_probesigs.gen_signal(probe)

        ## Instantiate traceport module
        self.num_probes = len(probes)
        self.trap_inst_ifc = SVAPI()
        trap_inst = ModuleInst(api=self.trap_inst_ifc, name='trace_port_gen')
        trap_inst.add_inputs(probes, connections=probes)
        trap_inst.add_input(scfg.emu_clk, connection=scfg.emu_clk)
        trap_inst.generate_instantiation()

        ## Assign probe signals via abs paths into design except for time probe
        self.assign_probesigs = SVAPI()
        for probe in scfg.digital_probes + scfg.analog_probes:
            self.assign_probesigs.assign_to(io_obj=probe, exp=probe.abs_path)

        ######################################################
        # Manage emulation clks Module
        ######################################################

        ## Instantiate all emulation clk signals
        self.inst_emu_clk_sigs = SVAPI()

        clk_io_pairs = []
        for derived_clk in scfg.clk_derived:
            if derived_clk.abspath_gated_clk is None:
                continue
            clk_val = digsig(f'clk_val_{derived_clk.name}')
            clk = digsig(f'clk_{derived_clk.name}')
            self.inst_emu_clk_sigs.gen_signal(clk_val)
            self.inst_emu_clk_sigs.gen_signal(clk)
            clk_io_pairs.append((clk_val, clk))

        # add default oscillator signals, in case default oscillator was used
        if self.use_default_oscillator:
            clk_val = digsig(f'clk_val_default_osc')
            clk = digsig(f'clk_default_osc')
            self.inst_emu_clk_sigs.gen_signal(clk_val)
            self.inst_emu_clk_sigs.gen_signal(clk)
            clk_io_pairs.append((clk_val, clk))

        ## Instantiate gen emulation clk module
        self.emu_clks_inst_ifc = SVAPI()

        emu_clks_inst = ModuleInst(api=self.emu_clks_inst_ifc, name="gen_emu_clks")
        emu_clks_inst.add_input(scfg.emu_clk_2x, connection=scfg.emu_clk_2x)
        emu_clks_inst.add_output(scfg.emu_clk, connection=scfg.emu_clk)

        for clk_val, clk in clk_io_pairs:
            emu_clks_inst.add_input(clk_val, connection=clk_val)
            emu_clks_inst.add_output(clk, connection=clk)

        emu_clks_inst.generate_instantiation()

        ######################################################
        # Manage default oscillator module
        ######################################################

        emu_dt = digsig('emu_dt', width=pcfg.cfg.dt_width)
        dt_req_default_osc = DigitalSignal(name=f'dt_req_default_osc', abspath='', width=pcfg.cfg.dt_width, signed=False)

        if scfg.use_default_oscillator:
            # instantiate API
            self.def_osc_api = SVAPI()

            # generate clock and reset signals

            # instantiate the default oscillator
            def_osc_inst = ModuleInst(
                api=self.def_osc_api,
                name='osc_model_anasymod',
                inst_name='def_osc_i'
            )

            emu_dt_req = digsig('emu_dt_req', width=pcfg.cfg.dt_width)

            def_osc_inst.add_output(scfg.emu_clk, connection=scfg.emu_clk)
            def_osc_inst.add_output(scfg.reset_ctrl, connection=scfg.reset_ctrl)
            def_osc_inst.add_output(emu_dt, connection=emu_dt)
            def_osc_inst.add_output(emu_dt_req, connection=dt_req_default_osc)
            def_osc_inst.add_output(digsig('cke'), connection=digsig('clk_val_default_osc'))
            def_osc_inst.generate_instantiation()
        else:
            self.def_osc_api = None

        ######################################################
        # Manage time manager Module
        ######################################################

        ## Instantiate all time manager signals
        self.inst_timemanager_sigs = SVAPI()

        self.inst_timemanager_sigs.gen_signal(emu_dt)

        dt_reqs = []
        for derived_clk in scfg.clk_derived:
            if derived_clk.abspath_dt_req is None:
                continue
            dt_req_sig = digsig(f'dt_req_{derived_clk.name}', width=pcfg.cfg.dt_width)
            self.inst_timemanager_sigs.gen_signal(dt_req_sig)
            dt_reqs.append(dt_req_sig)

        # add input for anasymod control dt request signal
        dt_req_stall = DigitalSignal(name=f'dt_req_stall', abspath='', width=pcfg.cfg.dt_width, signed=False)
        dt_reqs.append(dt_req_stall)

        # add input for dt request signal, in case a default oscillator is used
        if scfg.use_default_oscillator:
            dt_reqs.append(dt_req_default_osc)

        ## Instantiate time manager module
        self.time_manager_inst_ifc = SVAPI()
        time_manager_inst = ModuleInst(api=self.time_manager_inst_ifc, name='gen_time_manager')
        time_manager_inst.add_input(scfg.emu_clk, connection=scfg.emu_clk)
        time_manager_inst.add_input(scfg.reset_ctrl, connection=scfg.reset_ctrl)
        time_manager_inst.add_output(scfg.time_probe, scfg.time_probe)
        time_manager_inst.add_output(emu_dt, emu_dt)
        for dt_req_sig in dt_reqs:
            time_manager_inst.add_input(dt_req_sig, connection=dt_req_sig)

        time_manager_inst.generate_instantiation()

        ######################################################
        # Control module
        ######################################################

        self.ctrl_anasymod_inst_ifc = SVAPI()

        ctrl_anasymod_inst = ModuleInst(
            api=self.ctrl_anasymod_inst_ifc,
            name='ctrl_anasymod',
            inst_name='ctrl_anasymod_i'
        )

        ctrl_anasymod_inst.add_input(scfg.emu_ctrl_data, connection=scfg.emu_ctrl_data)
        ctrl_anasymod_inst.add_input(scfg.emu_ctrl_mode, connection=scfg.emu_ctrl_mode)
        ctrl_anasymod_inst.add_input(scfg.time_probe, connection=scfg.time_probe)
        ctrl_anasymod_inst.add_input(scfg.dec_thr_ctrl, connection=scfg.dec_thr_ctrl)
        ctrl_anasymod_inst.add_input(scfg.emu_clk, connection=scfg.emu_clk)
        ctrl_anasymod_inst.add_input(scfg.reset_ctrl, connection=scfg.reset_ctrl)
        ctrl_anasymod_inst.add_output(scfg.dec_cmp, connection=scfg.dec_cmp)
        ctrl_anasymod_inst.add_output(dt_req_stall, connection=dt_req_stall)

        ctrl_anasymod_inst.generate_instantiation()

        # indicate whether TSTOP_MSDSL should be used
        self.use_tstop = target.cfg.tstop is not None

        #####################################################
        # Instantiate testbench
        #####################################################
        self.tb_inst_ifc = SVAPI()
        for clk_indep_elem in scfg.clk_independent:
            self.tb_inst_ifc.gen_signal(clk_indep_elem)
        tb_inst = ModuleInst(api=self.tb_inst_ifc, name='tb')
        tb_inst.add_inputs(scfg.clk_independent, connections=scfg.clk_independent)
        tb_inst.generate_instantiation()

        #####################################################
        # CPU debug mode
        #####################################################
        self.dump_debug_signals = SVAPI()
        if pcfg.cfg.cpu_debug_mode:
            dbg_mods_list = pcfg.cfg.cpu_debug_hierarchies
            if isinstance(dbg_mods_list, list):
                for dbg_module in dbg_mods_list:
                    if isinstance(dbg_module, list):
                        self.dump_debug_signals.writeln(f'$dumpvars({dbg_module[0]}, {dbg_module[1]});')
                    elif isinstance(dbg_module, int) and len(dbg_mods_list) == 2: # only one single debug module was provided
                        self.dump_debug_signals.writeln(f'$dumpvars({dbg_mods_list[0]}, {dbg_mods_list[1]});')
                        break
                    else:
                        raise Exception(f'ERROR: unexpected format for cpu_debug_hierarchies attribute of '
                                        f'project_config: {dbg_mods_list}, expecting a list including <depth>, '
                                        f'<path_to_module> or a lists of such a list.')
            elif not dbg_mods_list:
                self.dump_debug_signals.writeln(f'$dumpvars(0, top.tb_i);')
            else:
                raise Exception(f'ERROR: unexpected format for cpu_debug_hierarchies attribute of '
                                f'project_config: {dbg_mods_list}, expecting tuple, list or None')

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

{{subst.plugin_includes.text}}

`default_nettype none

`ifndef SIMULATION_MODE_MSDSL
{{subst.module_ifc.text}}
`else
module top(
);
`endif // `ifndef SIMULATION_MODE_MSDSL

// Declaration of control signals
{{subst.inst_itl_ctlsigs.text}}
(* dont_touch = "true" *) logic [((`DT_WIDTH)-1):0] dt_req_stall;
{% if subst.use_default_oscillator %}
(* dont_touch = "true" *) logic [((`DT_WIDTH)-1):0] dt_req_default_osc;
{% endif %}

// Declaration of probe signals
{{subst.inst_probesigs.text}}

// create ext_clk signal when running in simulation mode
`ifdef SIMULATION_MODE_MSDSL
    logic ext_clk;
    {% for line in subst.clk_in_sim_sigs.text.splitlines() %}
        {{line}}
    {% endfor %}
`endif // `ifdef SIMULATION_MODE_MSDSL

// debug clk declaration
{{subst.dbg_clk_sigs.text}}

// emulation clock declarations
logic emu_clk, emu_clk_2x;

{% if subst.inst_timemanager_sigs is not none %}
// declarations for time manager
{{subst.inst_timemanager_sigs.text}}
{% endif %}

// declarations for emu clock generator
{{subst.inst_emu_clk_sigs.text}}

// instantiate testbench
{{subst.tb_inst_ifc.text}}

// Instantiation of control wrapper
{{subst.sim_ctrl_inst_ifc.text}}

{% if subst.num_probes !=0 %}
// Instantiation of traceport wrapper
{{subst.trap_inst_ifc.text}}
{% endif %}

// Clock generator
{{subst.clk_gen_ifc.text}}

{% if not subst.no_time_manager %}
// Emulation Clks Generator
{{subst.emu_clks_inst_ifc.text}}

{% if subst.def_osc_api is not none %}
// Default oscillator
{{subst.def_osc_api.text}}
{% endif %}

{% if subst.time_manager_inst_ifc.text is not none %}
// Time manager
{{subst.time_manager_inst_ifc.text}}
{% endif %}

// control signals
{{subst.ctrl_anasymod_inst_ifc.text}}
{% else %}
// This is a bit of a hack for now, to allow users to avoid
// using the time manager if they want to run at really
// high frequencies.  emu_clk will run *twice* as fast as
// specified because no_time_manager was set.
assign emu_clk = emu_clk_2x;
{% endif %}

// Assignment for derived clks
{{subst.derived_clk_assigns.text}}

// Assignment of custom control signals via absolute paths to design signals
{{subst.assign_custom_ctlsigs.text}}

{% if subst.num_probes !=0 %}
// Assignment of probe signals via absolute paths to design signals
{{subst.assign_probesigs.text}}
{% endif %}

// simulation control
`ifdef SIMULATION_MODE_MSDSL
    {% if subst.use_tstop %}
    // stop simulation after a predefined amount of emulation time
    localparam longint tstop_uint = (`TSTOP_MSDSL) / (`DT_SCALE);
    always @(emu_time) begin
        if (emu_time >= tstop_uint) begin
            $display("Ending simulation at emu_time=%e", emu_time*(`DT_SCALE));
            $finish;
        end
    end
    {% endif %}
    // dump waveforms to a specified VCD file
    `define ADD_QUOTES_TO_MACRO(macro) `"macro`"
    initial begin
        $dumpfile(`ADD_QUOTES_TO_MACRO({{subst.result_path_raw}}));
        
        // Signals to be dumped as well for debug purposes
        {{subst.dump_debug_signals.text}}
        
    end
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule

`default_nettype wire
'''

def main():
    pass

if __name__ == "__main__":
    main()



from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.util import back2fwd
from anasymod.targets import FPGATarget

class TemplEXECUTE_FPGA_SIM(JinjaTempl):
    def __init__(self, target: FPGATarget, start_time: float, stop_time: float, server_addr: str):
        '''
        :param start_time:  Point in time from which recording run data will start
        :param stop_time:   Point in time where FPGA run will be stopped
        :param server_addr: Hardware server address for hw server launched by Vivado
        '''

        super().__init__(trim_blocks=False, lstrip_blocks=False)
        pcfg = target.prj_cfg
        scfg = target.str_cfg

        # set server address
        self.server_addr = server_addr

        # set the paths to the BIT and LTX file
        self.bit_file = back2fwd(target.bitfile_path)
        self.ltx_file = back2fwd(target.ltxfile_path)

        # set the JTAG frequency.  sometimes it is useful to try a slower frequency than default if there
        # are problems with the debug hub clock
        self.jtag_freq = str(int(pcfg.cfg.jtag_freq))

        # set the "short" device name which is used to distinguish the FPGA part from other USB devices
        self.device_name = pcfg.board.short_part_name

        # set the path where the CSV file of results from the ILA should be written
        self.output = back2fwd(target.result_path_raw)
        self.ila_reset = target.prj_cfg.vivado_config.ila_reset
        #tbd remove vio_reset
        self.vio_reset = target.prj_cfg.vivado_config.vio_reset

        # set the window count to half the ILA depth.  this is required because the window depth will be "2"
        # unfortunately a window depth of "1" is not allowed in "BASIC" capture mode, which needed to allow
        # decimation of the sampling rate.
        self.window_count = str(pcfg.ila_depth/2)

        # calculate decimation ratio
        if stop_time is None:
            decimation_ratio_float = 2.0
        else:
            decimation_ratio_float = ((stop_time-start_time)/pcfg.cfg.dt)/(pcfg.ila_depth/2)

        # enforce a minimum decimation ratio of 2x since the window depth is 2
        decimation_ratio_float = max(decimation_ratio_float, 2.0)

        # we have to subtract 1 from decimation ratio due to the FPGA implementation of this feature
        # that is, a decimation ratio setting of "1" actually corresponds to a decimation factor of "2"; a setting
        # of "2" corresponds to a decimation ratio of "3", and so on.
        decimation_ratio_setting = int(round(decimation_ratio_float)) - 1

        # determine starting time as an integer value
        time = target.str_cfg.time_probe
        start_time_int = int(round(start_time/pcfg.cfg.dt_scale))

        # export the decimation ratio, starting time, and time signal name to the template
        self.time_name = time.name
        self.decimation_ratio_setting = str(decimation_ratio_setting)
        self.start_time_int = f"{int(time.width)}'u{start_time_int}"

        # set display radix for analog probes
        # TODO: is it necessary to wrap the commands in "catch"?
        self.analog_probe_radix = []
        for probe in scfg.analog_probes:
            self.analog_probe_radix += [
                f'catch {{set_property DISPLAY_RADIX SIGNED [get_hw_probes trace_port_gen_i/{probe.name}]}}'
            ]
        self.analog_probe_radix = '\n'.join(self.analog_probe_radix)

        # set display radix for digital probes
        # TODO: is it necessary to wrap the commands in "catch"?
        self.digital_probe_radix = []
        for probe in (scfg.digital_probes + [scfg.time_probe]):
            signed = 'SIGNED' if probe.signed else 'UNSIGNED'
            self.digital_probe_radix += [
                f'catch {{set_property DISPLAY_RADIX {signed} [get_hw_probes trace_port_gen_i/{probe.name}]}}'
            ]
        self.digital_probe_radix = '\n'.join(self.digital_probe_radix)

    TEMPLATE_TEXT = '''\
# Connect to hardware
open_hw
catch {disconnect_hw_server}
{% if subst.server_addr is none %}
connect_hw_server
{% else %}
connect_hw_server -url {{subst.server_addr}}
{% endif %}
set_property PARAM.FREQUENCY {{subst.jtag_freq}} [get_hw_targets]
open_hw_target

# Configure files to be programmed
set my_hw_device [get_hw_devices {{subst.device_name}}*]
current_hw_device $my_hw_device
refresh_hw_device $my_hw_device
set_property PROGRAM.FILE "{{subst.bit_file}}" $my_hw_device
set_property PROBES.FILE "{{subst.ltx_file}}" $my_hw_device
set_property FULL_PROBES.FILE "{{subst.ltx_file}}" $my_hw_device

# Program the device
program_hw_devices $my_hw_device
refresh_hw_device $my_hw_device

# VIO setup
set vio_0_i [get_hw_vios -of_objects $my_hw_device -filter {CELL_NAME=~"sim_ctrl_gen_i/vio_0_i"}]
set rst_hw_probe [get_hw_probes *rst* -of_objects $vio_0_i]

# ILA setup
set my_hw_ila [get_hw_ilas]
display_hw_ila_data [get_hw_ila_data hw_ila_data_1 -of_objects $my_hw_ila]

##################################################
# Code related to non-interactive mode starts here
##################################################

# Trigger setup
startgroup
set_property CONTROL.CAPTURE_MODE BASIC $my_hw_ila
set_property CONTROL.TRIGGER_POSITION 0 $my_hw_ila
endgroup
set_property TRIGGER_COMPARE_VALUE gt{{subst.start_time_int}} [get_hw_probes trace_port_gen_i/{{subst.time_name}} -of_objects $my_hw_ila]

# Capture setup

# depth should always be "2" because unfortunately "1" is not a valid option
set_property CONTROL.DATA_DEPTH 2 $my_hw_ila
set_property CONTROL.WINDOW_COUNT {{subst.window_count}} $my_hw_ila
set emu_dec_cmp_probe [get_hw_probes trace_port_gen_i/emu_dec_cmp -of_objects $my_hw_ila]
set_property CAPTURE_COMPARE_VALUE eq1'b1 $emu_dec_cmp_probe

# VIO: decimation ratio
set emu_dec_thr_vio [get_hw_probes sim_ctrl_gen_i/emu_dec_thr -of_objects $vio_0_i]
set_property OUTPUT_VALUE_RADIX UNSIGNED $emu_dec_thr_vio
set_property OUTPUT_VALUE {{subst.decimation_ratio_setting}} $emu_dec_thr_vio
commit_hw_vio $emu_dec_thr_vio

# Radix setup: analog probes
{{subst.analog_probe_radix}}

catch {{'{'}}set_property DISPLAY_RADIX SIGNED [get_hw_probes trace_port_gen_i/{{subst.time_name}}]{{'}'}}

# Radix setup: digital probes
{{subst.digital_probe_radix}}

# Put design in reset
startgroup
set_property OUTPUT_VALUE 1 $rst_hw_probe
commit_hw_vio $rst_hw_probe
endgroup

# Arm the ILA trigger
run_hw_ila $my_hw_ila

# Take design out of reset
startgroup
set_property OUTPUT_VALUE 0 $rst_hw_probe
commit_hw_vio $rst_hw_probe
endgroup

# Get data from the ILA
wait_on_hw_ila $my_hw_ila
display_hw_ila_data [upload_hw_ila_data $my_hw_ila]

# Write ILA data to a file
write_hw_ila_data -csv_file -force "{{subst.output}}" hw_ila_data_1
'''

def main():
    print(TemplEXECUTE_FPGA_SIM(cfg=EmuConfig()).render())

if __name__ == "__main__":
    main()



import pathlib
import re
import shutil
from pathlib import Path

from anasymod.util import call
from anasymod.generators.codegen import CodeGenerator
from anasymod.config import EmuConfig

class XSCTTCLGenerator(CodeGenerator):
    """
    Class for generating TCL scripts for XSCT
    """

    def __init__(self, pcfg: EmuConfig,xsct=None, version=None,
                 version_year=None, version_number=None,
                 xsct_install_dir=None):
        super().__init__()

        self.pcfg = pcfg
        self._xsct = self.pcfg.xsct_config.xsct

    def run(self, filename=r'run.tcl', err_str=None):
        # write the TCL script
        tcl_script = Path(self.pcfg.build_root).resolve() / filename
        self.write_to_file(tcl_script)

        # assemble the command
        cmd = [str(self.pcfg.xsct_config.xsct), str(tcl_script)]

        # run the script
        call(args=cmd, err_str=err_str)



import re
from anasymod.sim.sim import Simulator
from anasymod.generators.vivado import VivadoTCLGenerator

class VivadoSimulator(Simulator):
    def simulate(self):
        # set up the simulation commands
        v = VivadoTCLGenerator(target=self.target)

        # create a new project
        v.create_project(project_name=self.cfg.vivado_config.project_name,
                         project_directory=self.target.project_root,
                         force=True)

        # add all source files to the project (including header files)
        v.add_project_sources(content=self.target.content)

        # define the top module
        v.set_property('top', f"{{{self.target.cfg.top_module}}}", '[get_filesets {sim_1 sources_1}]')

        # set define variables
        v.add_project_defines(content=self.target.content, fileset='[get_filesets {sim_1 sources_1}]')

        # add include directories
        v.add_include_dirs(content=self.target.content, objects='[get_filesets {sim_1 sources_1}]')

        # if desired, treat Verilog (*.v) files as SystemVerilog (*.sv)
        if self.target.prj_cfg.cfg.treat_v_as_sv:
            v.writeln('set_property file_type SystemVerilog [get_files -filter {FILE_TYPE == Verilog}]')

        # launch the simulation
        v.set_property('{xsim.simulate.runtime}', '{-all}', '[get_fileset sim_1]')
        v.writeln('launch_simulation')

        # add any additional user-supplied flags
        for flag in self.flags:
            v.writeln(str(flag))

        # run the simulation
        v.run(filename='vivado_sim.tcl', err_str=re.compile('^(Error|Fatal):'))



class ConfigSections:
    """
    Container including enums for sections that can be used in config file.
    """
    PROJECT = "PROJECT"
    CPU_TARGET = "CPU_TARGET"
    FPGA_TARGET = "FPGA_TARGET"
    PLUGIN = "PLUGIN"
    STRUCTURE = "STRUCTURE"
    FPGASIM = "FPGASIM"


class BoardNames:
    """
    Container including enums for all supported boards.

    Currently supported boards are:
    PYNQ_Z1
    ARTY_A7
    VC707
    ZC702
    ZC706
    ZCU102
    ZCU106
    ZCU111
    ULTRA96
    TE0720

    """
    PYNQ_Z1 = 'PYNQ_Z1'
    ARTY_A7 = 'ARTY_A7'
    VC707 = 'VC707'
    ZC702 = 'ZC702'
    ZC706 = 'ZC706'
    ZCU102 = 'ZCU102'
    ZCU106 = 'ZCU106'
    ZCU111 = 'ZCU111'
    ULTRA96 = 'ULTRA96'
    TE0720 = 'TE0720'
    ARTY_200T_CUSTOM_LIDAR = 'ARTY_200T_CUSTOM_LIDAR'

class PortDir:
    """
    Container including enums for all supported port directions.
    """
    IN = "input"
    OUT = "output"
    INOUT = "inout"


class TraceUnitOperators:
    """
    Container including enums for all valid operators that can be used for setting up the trace unit
    """
    EQUAL = 'eq'
    NOTEQUAL = 'neq'
    GREATER = 'gt'
    GREATEREQUAL = 'gteq'
    LESSER = 'lt'
    LESSEREQUAL = 'lteq'

class FPGASimCtrl:
    """
    Container including enums for all supported FPGA control interfaces.
    """
    UART_ZYNQ = 'UART_ZYNQ'
    VIVADO_VIO = 'VIVADO_VIO'


class ResultFileTypes:
    """
    Container including enums for all supported data formats to store simulation/emulation results.
    """
    VCD = 'vcd'
    CSV = 'csv'


from anasymod.templates.templ import JinjaTempl
from anasymod.targets import FPGATarget
from anasymod.config import EmuConfig

class TemplExtClk(JinjaTempl):
    def __init__(self, target: FPGATarget):
        super().__init__(trim_blocks=True, lstrip_blocks=True)
        self.target = target

        self.ports = self.target.str_cfg.clk_i

        self.ext_clk_pin = self.target.prj_cfg.board.clk_pin
        self.ext_clk_io_std = self.target.prj_cfg.board.clk_io

        period_ns = 1e9 / self.target.prj_cfg.board.clk_freq
        self.ext_clk_full_period_ns = str(period_ns)
        self.ext_clk_half_period_ns = str(0.5*period_ns)

        self.zipped = zip(self.ext_clk_pin, [port.name for port in self.ports])

    TEMPLATE_TEXT = '''
# start auto-generated code for external clock

{% for boardport, designport in subst.zipped -%}
    set_property -dict {PACKAGE_PIN {{boardport}} IOSTANDARD {{subst.ext_clk_io_std}}} [get_ports {{designport}}]
{% endfor %}

create_clock -period {{subst.ext_clk_full_period_ns}} -name {{subst.ports[0].name}} -waveform {0.0 {{subst.ext_clk_half_period_ns}}} -add [get_ports {{subst.ports[0].name}}]
'''

def main():
    print(TemplExtClk(target=FPGATarget(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


class Define():
    def __init__(self, name, value=None, fileset=r"default"):
        self.name = name
        """ type(str) : mandatory setting; name of define. """

        self.value = value
        """ type(str) : value of define. """

        self.define = {self.name: self.value}

        self.fileset = fileset
        """ type(str) : Fileset, the source shall be associsted with. """


from anasymod.viewer.viewer import Viewer
from anasymod.util import call

class ScansionViewer(Viewer):
    def view(self):
        # build command
        #cmd = ['open', '/Applications/Scansion.app', self.target.cfg['vcd_path']]
        cmd = ['open', '/Applications/Scansion.app']

        # run command
        call(cmd)


import os

from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.util import back2fwd
from anasymod.targets import FPGATarget

class TemplPROBE_EXTRACT(JinjaTempl):
    def __init__(self, target: FPGATarget):
        super().__init__()
        self.project_dir = back2fwd(target.project_root)
        self.dcp_path = back2fwd(os.path.join(self.project_dir, target.prj_cfg.vivado_config.project_name + r".runs", r"synth_1", target.cfg.top_module + r".dcp"))

    TEMPLATE_TEXT = '''
# Load synthesized design
open_checkpoint "{{subst.dcp_path}}"

# Extract analog_signals
set ana_s [get_nets -hier -filter {analog_signal == "true"}]

regsub -all "\[{1}[0123456789]+\]{1}" $ana_s "" ana_s
set ana_s [lsort -unique $ana_s]

set ana_s_exponents [list]
set ana_s_widths [list]

foreach {signal} $ana_s {
	append signal [0]
	lappend ana_s_exponents [get_property fixed_point_exponent [get_nets $signal]]
	lappend ana_s_widths [get_property BUS_WIDTH [get_nets $signal]]
}

# Extract time_signal
set time_s [get_nets -hier -filter {time_signal == "true"}]
regsub -all "\[{1}[0123456789]+\]{1}" $time_s "" time_s
set time_s [lsort -unique $time_s]
set time_s_tmp $time_s
append time_s_tmp [0]

set time_s_exponent [get_property fixed_point_exponent [get_nets $time_s_tmp]]
set time_s_width [get_property BUS_WIDTH [get_nets $time_s_tmp]]

# Extract reset_signal
set reset_s [get_nets -hier -filter {reset_signal == "true"}]

# Extract other_signals marked for debug
set other_s [get_nets -hier -filter {digital_signal == "true" && MARK_DEBUG}]

set sb_s [list]
set sb_s [lsearch -regexp -not -inline -all $other_s "\[{1}[0123456789]+\]{1}"]

set mb_s [list]
set mb_s [lsearch -regexp -inline -all $other_s "\[{1}[0123456789]+\]{1}"]
regsub -all "\[{1}[0123456789]+\]{1}" $mb_s "" mb_s
set mb_s [lsort -unique $mb_s]

set mb_s_widths [list]

foreach {signal} $mb_s {
	append signal [0]
	lappend mb_s_widths [get_property BUS_WIDTH [get_nets $signal]]
}

set outputFile [open "{{subst.project_dir}}/probe_config.txt" w]
puts $outputFile [concat "ANALOG:" $ana_s]
puts $outputFile [concat "ANALOG_EXPONENT:" $ana_s_exponents]
puts $outputFile [concat "ANALOG_WIDTH:" $ana_s_widths]
puts $outputFile [concat "TIME:" $time_s]
puts $outputFile [concat "TIME_EXPONENT:" $time_s_exponent]
puts $outputFile [concat "TIME_WIDTH:" $time_s_width]
puts $outputFile [concat "RESET:" $reset_s]
puts $outputFile [concat "SB:" $sb_s]
puts $outputFile [concat "MB:" $mb_s]
puts $outputFile [concat "MB_WIDTH:" $mb_s_widths]
close $outputFile

close_design
'''

def main():
    print(TemplPROBE_EXTRACT(cfg=EmuConfig()).render())

if __name__ == "__main__":
    main()


# Terminal formatting codes adapted from http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#8-colors

import os

COLOR_CODES = dict(
    black = '\x1b[30m',
    red = '\x1b[31m',
    green = '\x1b[32m',
    yellow = '\x1b[33m',
    blue = '\x1b[34m',
    magenta = '\x1b[35m',
    cyan = '\x1b[36m',
    white = '\x1b[37m',
    bright = '\x1b[1m',
    reset = '\x1b[0m',
)

def cprint(s, *args, color=None, bold=None, **kwargs):
    # set defaults
    if bold is None:
        bold = color is not None

    # build up the string to print
    to_print = f'{s}'
    if bold:
        to_print = COLOR_CODES['bright']  + to_print
    if color is not None:
        to_print = COLOR_CODES[color.lower()] + to_print
    if bold or color is not None:
        to_print += COLOR_CODES['reset']

    # print the string
    print(to_print, *args, **kwargs)

def cprint_announce(title, text, color=None, bold=None):
    if os.name != 'nt':
        cprint(title, color=color, bold=bold, end='')
    else:
        print(title)
    print(text)

def cprint_block_start(title, color=None, bold=None):
    if os.name != 'nt':
        cprint(f'<{title}>', color=color, bold=bold)
    else:
        print(f'<{title}>')

def cprint_block_end(title, color=None, bold=None):
    if os.name != 'nt':
        cprint(f'</{title}>', color=color, bold=bold)
    else:
        print(f'<{title}>')

def cprint_block(str_list, title, color=None, bold=None):
    # don't print out anything if the string list has zero length
    if len(str_list) == 0:
        return

    # otherwise print out the text with the start and end indicated
    cprint_block_start(title, color=color, bold=bold)
    for s in str_list:
        print(str(s).rstrip())
    cprint_block_end(title, color=color, bold=bold)


import serial, os
import serial.tools.list_ports as ports
from .console_print import cprint_block
from pathlib import Path
from anasymod.base_config import BaseConfig
from anasymod.sim_ctrl.ctrlapi import CtrlApi
from anasymod.enums import ConfigSections
from anasymod.config import EmuConfig
from anasymod.structures.structure_config import StructureConfig
from anasymod.emu.xsct_emu import XSCTEmulation
from anasymod.generators.gen_api import CodeGenerator
from anasymod.templates.launch_ILA_tcl import TemplLAUNCH_ILA_TCL
from anasymod.enums import TraceUnitOperators
from anasymod.sim_ctrl.datatypes import AnalogProbe, DigitalSignal
from anasymod.util import expand_path
from anasymod.wave import ConvertWaveform
from anasymod.files import mkdir_p


class UARTCtrlApi(CtrlApi):
    """
    Start an interactive control interface to HW target for running regression tests or design exploration/debug.
    For FPGA/Emulators, as a pre-requisit, bitstream must have been created and programmed. Additionally any eSW
    necessary in the targeted system must have already been programmed.
    """
    def __init__(self, result_path_raw, result_type_raw, result_path, prj_cfg: EmuConfig, scfg: StructureConfig,
                 content, project_root, ltxfile_path, top_module, cwd=None, err_strs=None, debug=False,
                 float_type=False, prompt='Vivado% '):
        super().__init__(cwd=cwd, pcfg=prj_cfg, scfg=scfg, prompt=prompt, debug=debug)
        # set defaults
        if err_strs is None:
            err_strs = ['ERROR', 'FATAL']

        self.content = content
        self.project_root = project_root
        self.top_module = top_module

        self.result_path_raw = result_path_raw
        self.result_type_raw = result_type_raw
        self.result_path = result_path
        self.float_type = float_type

        self.debug = debug
        self.err_strs = err_strs

        self.tcl_console_running = False
        self.server_addr = None
        self.ltxfile_path = ltxfile_path

        # Initialize control config
        self.cfg = Config(cfg_file=prj_cfg.cfg_file)

        vid = self.pcfg.board.uart_zynq_vid
        if isinstance(vid, list):
            self.vid_list = vid
        else:
            self.vid_list = [vid]

        pid = self.pcfg.board.uart_zynq_pid
        if isinstance(pid, list):
            self.pid_list = pid
        else:
            self.pid_list = [pid]

        self.uart_suffix = self.pcfg.board.uart_suffix

        self.port_list = []
    ### User Functions

    def sendline(self, line, timeout=float('inf')):
        """
        Send a single line in target shell specific language e.g. in tcl for tcl shell.
        :param line: Line that shall be send to/processed by shell
        :param timeout: Maximum time granted for operation to finish
        :return: Return string from Vivado TCL interpreter
        """
        if self.debug:
            cprint_block([line], title='SEND', color='magenta')

        self.proc.sendline(line)
        before = self._expect_prompt(timeout=timeout)

        # make sure that there were no errors
        for err_str in self.err_strs:
            if err_str in before:
                raise Exception(f'Found {err_str} in output from Vivado.')

    def source(self, script, timeout=float('inf')):
        """
        Source a script written language for targeted shell.
        :param script: Name/Path to script that shall be sourced
        :param timeout: Maximum time granted for operation to finish
        """
        script = Path(script).resolve()
        res = self.sendline(f'source {script.as_posix()}', timeout=timeout)

    def setup_trace_unit(self, trigger_name, trigger_operator, trigger_value, sample_decimation=None, sample_count=None):
        """
        Setup the trace unit. This involves defining the signal, that shall be used to start tracing and defining the
        comparison operator, which is used to monitor the trigger signal. Also arm the trace unit at the end.

        Note: In case the trigger value is set too high and an overflow occurs, the simulation will hang, as the trigger
        condition will never be met.

        :param trigger_name:        Probe signal to be used as trigger signal.
        :param trigger_operator:    Comparison operator that is used when monitoring the trigger signal to detect,
                                    when a trigger shall be fired. Available oerators are:
                                        eq (equal),
                                        neq (not equal),
                                        gt (greater than),
                                        gteq (greater or equal than),
                                        lt (lesser than),
                                        lteq (lesser or equal than)
        :param trigger_value:       Value, the probe signal is compared to.
        :param sample_decimation:   Number of samples to be skipped during recording. By default, every sample will
                                    stored in the results file, adding a sample_decimation will allow to only record
                                    every x sample.
        :param sample_count:        Number of samples to be recorded. This number shall not exceed that maximum ILA
                                    depth defined during project setup AND needs to be a multiple of 2.
        """

        # Start vivado TCL interpreter, if it is not already running
        if not self.tcl_console_running:
            super()._initialize_vivado_tcl()
            self._setup_ila_ctrl(server_addr=self.server_addr)
            self.tcl_console_running = True

        # Compute timeout, that needs to be set for ila recording. Due to a bug in the ILA core, recording does not
        # return, unless a timeout is specified, for recordings exceeding 0.26 s
        self.record_timeout = 0.0

        trigger_obj = None

        if sample_count:
            # This solution I got from https://stackoverflow.com/questions/57025836/check-if-a-given-number-is-power-of-two-in-python
            # and is checking if sample_count is a power of two using bit manipulations
            if (sample_count & (sample_count-1) == 0) and sample_count != 0:
                depth = sample_count
            else:
                raise Exception(f'ERROR: Sample count needs to be a power of 2, but is set to: {sample_count}')
        else:
            depth = self.pcfg.ila_depth

        # Check, if provided trigger_signal is a valid signal connected to the ILA
        for probe in [self.scfg.time_probe] + self.scfg.analog_probes + self.scfg.digital_probes:
            if trigger_name == probe.name:
                trigger_obj = probe
            elif trigger_name == 'time':
                trigger_obj = self.scfg.time_probe
            else:
                raise Exception(f'ERROR: Provided trigger_signal:{trigger_name} is not a valid probe in the existing project!')

        # Check if operator set is legal
        if not trigger_operator in TraceUnitOperators.__dict__.values():
            raise Exception(f'ERROR: Provided operator:{trigger_operator} is not legal! Supported operators are:{TraceUnitOperators.__dict__.values()}')

        # Convert compare value depending on signal type
        if isinstance(trigger_obj, AnalogProbe):
            if not isinstance(trigger_value, float):
                raise Exception(
                    f'ERROR: trigger_value type for an analog trigger signal shall ne float, it is instead:{type(trigger_value)}')
            value_int = int(round(trigger_value * (2 ** (int(-trigger_obj.exponent)))))
            trigger_value_as_bin = f"{int(trigger_obj.width)}'u{value_int}"
        elif isinstance(trigger_obj, DigitalSignal):
            # For the time signal, conversion from float to int is necessary
            if trigger_name in ['time', self.scfg.time_probe.name]:
                # Convert trigger value to integer considering dt_scale
                trigger_value_as_int = int(float(trigger_value) / float(self.pcfg.cfg.dt_scale))
                # Represent as binary and expand to time_width
                trigger_value_as_bin = f"{trigger_obj.width}'b{bin(trigger_value_as_int).replace('b', '').zfill(self.pcfg.cfg.time_width)}"
                # Extend record_timeout according to time trigger
                self.record_timeout += float(trigger_value)
            elif isinstance(trigger_value, int):
                trigger_value_as_bin = f"{trigger_obj.width}'u{trigger_value}"
            else:
                p = set(trigger_value)
                s = {'0', '1'}

                if s == p or p == {'0'} or p == {'1'}:
                    trigger_value_as_bin = f"{trigger_obj.width}'b{trigger_value}"
                else:
                    raise Exception(f'ERROR: Provided trigger value:{trigger_value} is not valid for a digital trigger '
                                    f'signal, either provide and interger or binary string!')
        else:
            raise Exception(f'ERROR: No valid signal type for provided trigger signal:{trigger_name} Type:{type(trigger_obj)}')

        self.sendline(f'set_property CONTROL.CAPTURE_MODE BASIC $ila_0_i')
        self.sendline(f'set_property CONTROL.TRIGGER_POSITION 0 $ila_0_i')
        self.sendline(f"set_property TRIGGER_COMPARE_VALUE {trigger_operator}{trigger_value_as_bin} [get_hw_probes trace_port_gen_i/{trigger_obj.name} -of_objects $ila_0_i]")

        # Data depth is set to 2, as 1 is not supported by Vivado's ILA Core
        self.sendline(f'set_property CONTROL.DATA_DEPTH {depth} $ila_0_i')

        # Window count is set to half of the selected ILA depth
        #self.sendline(f'set_property CONTROL.WINDOW_COUNT {int(window_count)} $ila_0_i')
        self.sendline(f'set_property CONTROL.WINDOW_COUNT 1 $ila_0_i')
        self.sendline(f"set_property CAPTURE_COMPARE_VALUE eq1'b1 [get_hw_probes trace_port_gen_i/emu_dec_cmp -of_objects $ila_0_i]")

        # Set decimation threshold signal to value defined in sample_decimation if set
        if sample_decimation:
            if isinstance(sample_decimation, int): # Check if sample_decimation is an integer
                self.set_param(name=self.scfg.dec_thr_ctrl.name, value=sample_decimation, timeout=30)
            else:
                raise Exception(f'Provided sample_decimation value is of wrong type, expecting integer and got:{type(sample_decimation)}')
        else:
            sample_decimation = 0

        self.arm_trace_unit()

        # conclude calculation of record_timeout by considering time needed to collect samples and representing
        # over minutes

        self.record_timeout += float(self.pcfg.cfg.dt) * depth * sample_decimation
        self.record_timeout = self.record_timeout / 60

    def arm_trace_unit(self):
        """
        Arm the trace unit, this will delete the buffer and arm the trigger.
        """
        self.sendline('run_hw_ila $ila_0_i')

    def wait_on_and_dump_trace(self, result_file=None):
        """
        Wait until the trace unit stopped recording data. Transmit this data to the host PC, store by default to the raw
        result file path, or a custom path provided by the user, and convert analog values from fixed-point to float.
        Finally store it to a .vcd file in the default location.

        :param result_file: Optionally, it is possible to provide a custom result file path.
        """

        if result_file is not None:
            # Expand provided path, paths relative to project root are also supported
            result_path = expand_path(result_file, rel_path_reference=self.pcfg.root)

            # Create raw result path by adding _raw to the filename
            result_path_raw = os.path.join(os.path.dirname(result_path),
                                           os.path.basename(os.path.splitext(result_path)[0]) + '_raw' +
                                           os.path.splitext(result_path)[1])
            print(f'Simulation results will be stored in:{result_path}')
        else:
            result_path = self.result_path
            result_path_raw = self.result_path_raw

        if not result_path:
            raise Exception(f'ERROR: provided result_file:{result_file} is not valid!')

        # wait until trace buffer is full
        self.sendline(f'wait_on_hw_ila -timeout {self.record_timeout} $ila_0_i')

        # transmit and dump trace buffer data to a CSV file
        self.sendline('current_hw_ila_data [upload_hw_ila_data $ila_0_i]')

        if not os.path.isdir(os.path.dirname(result_path_raw)):
            mkdir_p(os.path.dirname(result_path_raw))

        self.sendline(f'write_hw_ila_data -csv_file -force {{{result_path_raw}}} [current_hw_ila_data]')

        # Convert to .vcd and from fixed-point to float
        ConvertWaveform(result_path_raw=result_path_raw,
                        result_type_raw=self.result_type_raw,
                        result_path=result_path,
                        str_cfg=self.scfg,
                        float_type=self.float_type,
                        dt_scale=self.pcfg.cfg.dt_scale)

    def get_param(self, name, timeout=30):
        """
        Read value of a control parameter in design.
        :param name: Name of control parameter to be read
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        self._write(name=self.cfg.get_operation_prefix+name)
        return self._read()

    def set_param(self, name, value, timeout=30):
        """
        Set value of a control parameter in design.
        :param name: Name of control parameter to be set
        :param value: Value of control parameter sto be set
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        self._write(name=self.cfg.set_operation_prefix+name, value=value)
        if self._read():
            raise Exception(f"ERROR: Couldn't properly write: {self.cfg.set_operation_prefix+name}={value} command to FPGA.")

    def set_reset(self, value, timeout=30):
        """
        Control the 'emu_rst' signal, in order to put the system running on the FPGA into or out of reset state.
        :param value: Value of reset signal, 1 will set it to reset and 0 will release reset.
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_param(name= self.scfg.reset_ctrl.name, value=value, timeout=timeout)

    def get_emu_time_int(self, timeout=30):
        """
        Get current time of the FPGA simulation as an unscaled integer value.
        :param timeout: Maximum time granted for operation to finish
        """
        emu_time_vio = self.get_param(name=self.scfg.emu_time_vio.name, timeout=timeout)
        return int(emu_time_vio)

    def set_ctrl_mode(self, value, timeout=30):
        """
        Set the control mode that shall be applied to stall the FPGA simulation.
        :param value: Integer value setting the currently active control mode:
                        0:  FPGA simulation is not stalled
                        1:  FPGA simulation is immediately stalled
                        2:  FPGA simulation is stalled, after time stored in *ctrl_data* has passed, starting from
                            the point in time this mode has been selected.
                        3:  FPGA simulation is stalled, once time value stored in *ctrl_data* has been reached
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_param(name=self.scfg.emu_ctrl_mode.name, value=value, timeout=timeout)

    ### Utility Functions

    def _initialize(self):
        """
        Initialize the control interface, this is usually done after the bitstream was programmed successfully on the FPGA.
        :return:
        """
        if self.cfg.comport is None:  # run auto search for finding the correct COM port
            # find all available COM ports
            comports = [port for port in ports.comports(include_links=True)]
            # check if any COM port is compliant to known vids and pids and if so store the device_id
            for port in comports:
                if ((port.vid in self.vid_list) and (port.pid in self.pid_list)):
                    if (self.uart_suffix is None) or (port.location.endswith(self.uart_suffix)):
                        self.port_list.append(port.device)

            for port in self.port_list:
                try:
                    self.ctrl_handler = serial.Serial(port, int(self.cfg.baud_rate), timeout=self.cfg.timeout,
                                                      parity=self.cfg.parity, stopbits=self.cfg.stopbits,
                                                      bytesize=self.cfg.bytesize)
                    self.cfg.comport = port
                except:
                    pass
            if self.cfg.comport is None:
                raise Exception(
                    f"ERROR: No COM port could be opened to enable connection to FPGA, found ports were: {self.port_list}.")

        else:
            try:
                self.ctrl_handler = serial.Serial(self.cfg.comport, int(self.cfg.baud_rate), timeout=self.cfg.timeout,
                                                  parity=self.cfg.parity, stopbits=self.cfg.stopbits,
                                                  bytesize=self.cfg.bytesize)
            except:
                raise Exception(f"ERROR: Provided COM port: {self.cfg.comport} could not ne opened for communication.")

    def _setup_ctrl(self, server_addr, *args, **kwargs):
        """
        Prepare instrumentation on the FPGA to allow interactive control.
        :param server_addr: Address of remote hardware server
        :return:
        """

        self.server_addr = server_addr

        # program the firmware
        XSCTEmulation(pcfg=self.pcfg,
                      content=self.content,
                      project_root=self.project_root,
                      top_module=self.top_module
                      ).program(server_addr=server_addr, *args, **kwargs)

    def _write(self, name, value=None):
        # check is space is in any of the give input strings
        if ' ' in [name, value]:
            raise Exception(f"Blanks in any of the provided argument strings;{name}, {value}; sent via control interface are not allowed!")
        if value is not None:
            self.ctrl_handler.write((f'{str(name)} {str(value)}\n'.encode('utf-8')))
        else:
            self.ctrl_handler.write((f'{str(name)}\n'.encode('utf-8')))
        self.ctrl_handler.flush()

    def _read(self, count=1):
        for idx in range(count):
            result = self.ctrl_handler.readline().decode('utf-8').rstrip()

            if result not in ['', None]:
                return int(result)
        raise Exception(f"ERROR: Couldn't read from FPGA after:{count} attempts.")

    def _setup_ila_ctrl(self, server_addr):
        """
        Prepare ILA instrumentation on the FPGA to allow interactive control.
        :param server_addr: Address of remote hardware server
        :return:
        """
        launch_script = os.path.join(os.path.dirname(os.path.dirname(self.result_path_raw)), r"launch_FPGA.tcl")
        codegen = CodeGenerator()
        codegen.use_templ(TemplLAUNCH_ILA_TCL(pcfg=self.pcfg, scfg=self.scfg,
                                               ltxfile_path=self.ltxfile_path, server_addr=server_addr))
        codegen.write_to_file(launch_script)
        self.source(script=launch_script)

    def __del__(self):
        """
        Close connection to shell.
        :return:
        """
        pass

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """
    def __init__(self, cfg_file):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.FPGASIM)

        self.comport = None
        """ type(str) : Name of the COM port, that is attached to FPGA """

        self.baud_rate = 115200
        """ type(int): Baudrate used for communication with UART on FPGA """

        self.timeout = None
        """ type(int): Timeout in seconds used when waiting for a response during data transmission """

        self.parity = serial.PARITY_NONE
        """ type(str): Parity mode used for UART communication """

        self.stopbits=1
        """ type(int): Number of stop bits used when transmitting a package """

        self.bytesize=serial.EIGHTBITS
        """ type(int): Number of bits per transmitted package """

        self.set_operation_prefix='SET_'
        """ type(str): Default operator prefix added to every set operation."""

        self.get_operation_prefix = 'GET_'
        """ type(str): Default operator prefix added to every get operation."""

def main():
   ctrl = UARTCtrlApi(prj_cfg=EmuConfig(root='',cfg_file=''))
   ctrl.set_param(name=0, value=3)
   ctrl.set_param(name=1, value=4)
   print(ctrl.get_param(name=0))
   print(ctrl.get_param(name=1))

if __name__ == "__main__":
   main()


from anasymod.templates.templ import JinjaTempl
from anasymod.util import back2fwd
from anasymod.generators.gen_api import SVAPI
from anasymod.config import EmuConfig
from anasymod.structures.structure_config import StructureConfig

class TemplLAUNCH_FPGA_SIM(JinjaTempl):
    def __init__(self, pcfg: EmuConfig, scfg: StructureConfig, bitfile_path, ltxfile_path, server_addr: str):
        super().__init__(trim_blocks=False, lstrip_blocks=False)
        pcfg = pcfg
        scfg = scfg

        # set server address
        self.server_addr = server_addr

        # set the paths to the BIT and LTX file
        self.bit_file = back2fwd(bitfile_path)
        self.ltx_file = back2fwd(ltxfile_path)

        # set the JTAG frequency.  sometimes it is useful to try a slower frequency than default if there
        # are problems with the debug hub clock
        self.jtag_freq = str(int(pcfg.cfg.jtag_freq))

        # save some parameters from the board configuration
        self.device_name = pcfg.board.short_part_name
        self.no_rev_check = pcfg.board.no_rev_check

        # Set aliases for VIOs
        self.ctrl_io_aliases = SVAPI()
        for io in scfg.digital_ctrl_inputs + scfg.digital_ctrl_outputs + \
                  scfg.analog_ctrl_inputs + scfg.analog_ctrl_outputs:
            self.ctrl_io_aliases.writeln(f'set {io.name} [get_hw_probes "sim_ctrl_gen_i/{io.name}" -of_objects $vio_0_i]')

        # Set radix for VIOs
        self.ctrl_io_radix = SVAPI()
        for digital_in in scfg.digital_ctrl_inputs:
            signed = 'SIGNED' if digital_in.signed else 'UNSIGNED'
            self.ctrl_io_radix.writeln(f'set_property OUTPUT_VALUE_RADIX {signed} ${digital_in.name}')

        for digital_out in scfg.digital_ctrl_outputs:
            signed = 'SIGNED' if digital_out.signed else 'UNSIGNED'
            self.ctrl_io_radix.writeln(f'set_property INPUT_VALUE_RADIX {signed} ${digital_out.name}')

        for analog_in in scfg.analog_ctrl_inputs:
            self.ctrl_io_radix.writeln(f'set_property OUTPUT_VALUE_RADIX SIGNED ${analog_in.name}')

        for analog_out in scfg.analog_ctrl_outputs:
            self.ctrl_io_radix.writeln(f'set_property INPUT_VALUE_RADIX SIGNED ${analog_out.name}')

        # Set aliases for probes
        self.probe_aliases = SVAPI()
        for probe in scfg.digital_probes + scfg.analog_probes + [scfg.time_probe]:
            self.probe_aliases.writeln(f'set {probe.name} [get_hw_probes "trace_port_gen_i/{probe.name}" -of_objects $ila_0_i]')

        # Set radix for probes
        self.probe_radix = SVAPI()
        for digital_probe in (scfg.digital_probes + [scfg.time_probe]):
            signed = 'SIGNED' if digital_probe.signed else 'UNSIGNED'
            self.probe_radix.writeln(f'set_property DISPLAY_RADIX {signed} ${digital_probe.name}')

        for analog_probe in scfg.analog_probes:
            self.probe_radix.writeln(f'set_property DISPLAY_RADIX SIGNED ${analog_probe.name}')

    TEMPLATE_TEXT = '''\
# Connect to hardware
open_hw
catch {disconnect_hw_server}
{% if subst.server_addr is none %}
connect_hw_server
{% else %}
connect_hw_server -url {{subst.server_addr}}
{% endif %}
set_property PARAM.FREQUENCY {{subst.jtag_freq}} [get_hw_targets]
open_hw_target

{% if subst.no_rev_check %}
# Don't check revision compatibility between the bitstream
# and FPGA.  This is needed for some boards that use
# engineering samples.
set_param xicom.use_bitstream_version_check false 
{% endif %}

# Configure files to be programmed
set my_hw_device [get_hw_devices {{subst.device_name}}*]
current_hw_device $my_hw_device
refresh_hw_device $my_hw_device
set_property PROGRAM.FILE "{{subst.bit_file}}" $my_hw_device
set_property PROBES.FILE "{{subst.ltx_file}}" $my_hw_device
set_property FULL_PROBES.FILE "{{subst.ltx_file}}" $my_hw_device

# Program the device
program_hw_devices $my_hw_device
refresh_hw_device $my_hw_device

# VIO setup
set vio_0_i [get_hw_vios -of_objects $my_hw_device -filter {CELL_NAME=~"sim_ctrl_gen_i/vio_0_i"}]
set rst_hw_probe [get_hw_probes *rst* -of_objects $vio_0_i]

# ILA setup
set ila_0_i [get_hw_ilas -of_objects $my_hw_device -filter {CELL_NAME=~"trace_port_gen_i/ila_0_i"}]

##############################################
# Code related to interactive mode starts here
##############################################

# configure VIO for low latency
set_property CORE_REFRESH_RATE_MS 0 $vio_0_i

# set aliases to VIOs
{{subst.ctrl_io_aliases.text}}

# configure VIO radix
{{subst.ctrl_io_radix.text}}

# configure the ILA for low latency
set_property CORE_REFRESH_RATE_MS 0 $ila_0_i

# set aliases to ILA probes
{{subst.probe_aliases.text}}

# configure radix for ILA probes
{{subst.probe_radix.text}}
'''

def main():
    print(TemplLAUNCH_FPGA_SIM(target=FPGATarget(prj_cfg=EmuConfig(root='', cfg_file=''), plugins=[], name=r"test"), server_addr='').render())

if __name__ == "__main__":
    main()



import numpy as np
import scipy.interpolate

class Waveform:
    def __init__(self, data, time):
        self.data = np.array(data, dtype=float)
        self.time = np.array(time, dtype=float)
        self.interp = scipy.interpolate.interp1d(
            time, data,
            bounds_error=False, fill_value=(data[0], data[-1])
        )

        # attach measurement objects in a way that is compatible with pverify
        measurements = Measurements(self)
        self.Measurements_Base = measurements
        self.Measurements_NonPeriodic = measurements
        self.Measurements_Periodic = measurements

    def check_in_limits(self, wave_lo, wave_hi):
        # compute bounds
        lo_bnds = wave_lo.interp(self.time)
        hi_bnds = wave_hi.interp(self.time)

        # uncomment for graphical debugging
        # import matplotlib.pyplot as plt
        # plt.plot(self.time, lo_bnds)
        # plt.plot(self.time, hi_bnds)
        # plt.plot(self.time, self.data)
        # plt.legend(['lo_bnds', 'hi_bnds', 'data'])
        # plt.show()

        # find any point where the waveform is out of spec
        in_spec = np.logical_and(lo_bnds <= self.data, self.data <= hi_bnds)
        if np.all(in_spec):
            pass
        else:
            indices = np.where(np.logical_not(in_spec))
            times = self.time[indices]
            raise Exception(f'Waveform is out of spec.  Check times: {times}')

class Measurements:
    def __init__(self, wave: Waveform):
        self.wave = wave

    def min(self):
        return np.min(self.wave.data)

    def max(self):
        return np.max(self.wave.data)

    def peak_to_peak(self):
        return self.max() - self.min()

    def find_settled(self, lo, hi):
        # find first index where the waveform is in the settled range
        idx = np.argmax((lo <= self.wave.data) & (self.wave.data <= hi))

        # if there is no such value, idx will be zero, so we have to check
        # for that condition
        if not (lo <= self.wave.data[idx] <= hi):
            raise Exception('Waveform did not settle to the specified range.')

        return self.wave.time[idx]

    def frequency(self, level=0, slope='rise', hysteresis=0):
        # find the crossing times
        crossings = []
        armed = False
        for t, v in zip(self.wave.time, self.wave.data):
            if slope == 'rise':
                if armed and (v > (level + hysteresis)):
                    crossings.append(t)
                    armed = False
                elif v < (level - hysteresis):
                    armed = True
            elif slope == 'fall':
                if armed and (v < (level - hysteresis)):
                    crossings.append(t)
                    armed = False
                elif v > (level + hysteresis):
                    armed = True
            else:
                raise Exception(f"Unknown slope type: {slope}")

        # measure time from the first crossing to the last crossing,
        # as well as the number of periods
        dt = crossings[-1] - crossings[0]
        num = len(crossings) - 1

        # return the average frequency
        return num/dt

    def frequency_average(self, *args, **kwargs):
        # not sure what this function is supposed to do as compared to the previous
        # one, so I'm leaving this as an alias to "frequency" for now
        return self.frequency(*args, **kwargs)



import os
import re

from anasymod.sim.sim import Simulator
from anasymod.util import call

class IcarusSimulator(Simulator):
    def compile(self):
        # build up the simulation command
        cmd = [self.cfg.icarus_config.iverilog, '-g2012', '-o', self.cfg.icarus_config.output_file_path, '-s',
               self.target.cfg.top_module]

        # if desired, only preprocess
        if self.cfg.cfg.preprocess_only:
            cmd.append('-E')

        # add defines
        for define in self.target.content.defines:
            for k, v in define.define.items():
                if v is not None:
                    cmd.extend(['-D', f"{k}={v}"])
                else:
                    cmd.extend(['-D', f"{k}"])

        # add include directories, remove filename from paths and create a list of inc dirs removing duplicates
        inc_dirs = set()

        for sources in self.target.content.verilog_headers:
            for src in sources.files:
                inc_dirs.add(os.path.dirname(src))

        for sources in self.target.content.include_dirs:
            for src in sources.files:
                inc_dirs.add(src)

        for inc_dir in inc_dirs:
            cmd.extend(['-I', inc_dir])

        # add verilog source files
        for sources in self.target.content.verilog_sources:
            for src in sources.files:
                cmd.append(src)

        # add HDL files for functional models
        for sources in self.target.content.functional_models:
            for src in sources.gen_files:
                cmd.append(src)

        # add any additional user-supplied flags
        for flag in self.flags:
            cmd.append(str(flag))

        # run iverilog
        call(cmd, cwd=self.cfg.build_root)

    def run(self):
        args = [self.cfg.icarus_config.vvp, self.cfg.icarus_config.output_file_path]
        cwd = self.cfg.build_root
        err_str = re.compile('^(ERROR|FATAL):')
        call(args, cwd=cwd, err_str=err_str)

    def simulate(self):
        self.compile()

        if not self.cfg.cfg.preprocess_only:
            self.run()



import os
import platform

from collections import OrderedDict

from anasymod.sim.sim import Simulator
from anasymod.util import call
from anasymod.targets import CPUTarget
from anasymod.config import EmuConfig

class XceliumSimulator(Simulator):
    def __init__(self, target: CPUTarget, flags=None):
        super().__init__(target=target, flags=flags)
        self.unit = None
        self.id = None

    def simulate(self, licqueue=True, smartorder=True, timescale='1ns/1ps'):
        # prepare ifxxcelium makefile by adding inicio target
        self.prepare()

        # build up the simulation command
        cmd = []
        cmd += [self.cfg.xcelium_config.xrun]
        if "ifxxcelium" in self.cfg.xcelium_config.xrun:
            cmd += self.target.prj_cfg.xcelium_config.lsf_opts.split()
            cmd += ['inicio']
            if self.unit:
                cmd += ["-unit", self.unit]
            if self.id:
                cmd += ["-id", self.id]
            cmd += ['--']

        cmd += ['-top', self.target.cfg.top_module]
        cmd += ['-input', self.cfg.xcelium_config.tcl_input_path]

        # order independent compilation for VHDL
        if smartorder:
            cmd += ['-smartorder']

        # wait to get a license
        if licqueue:
            cmd += ['-licqueue']

        # specify the default timescale
        if timescale is not None:
            cmd += ['-timescale', f'{timescale}']

        # 64-bit or 32-bit mode
        # TODO: is this actually necessary?  the problem is that sometimes xrun is submitted through a bsub command,
        # so the remote machine architecture isn't necessarily the same as that of the local machine that submits
        # the job.
        if '64bit' in platform.architecture():
            cmd += ['-64bit']

        # add defines
        for define in self.target.content.defines:
            for k, v in define.define.items():
                if v is not None:
                    cmd.append(f"+define+{k}={v}")
                else:
                    cmd.append(f"+define+{k}")

        # add include directories, remove filename from paths and create a list of inc dirs removing duplicates
        inc_dirs = set()

        for sources in self.target.content.verilog_headers:
            for src in sources.files:
                inc_dirs.add(os.path.dirname(src))

        for sources in self.target.content.include_dirs:
            for src in sources.files:
                inc_dirs.add(src)

        for inc_dir in inc_dirs:
            cmd.extend(['-incdir', inc_dir])

        # add Verilog source files
        for sources in self.target.content.verilog_sources:
            for src in sources.files:
                cmd.append(src)

        # add HDL sources for functional models
        for sources in self.target.content.functional_models:
            for src in sources.gen_files:
                cmd.append(src)

        # add VHDL source files
        # TODO: is it actually necessary to consolidate makelib commands together?  if subsequent makelib commands
        # append rather than replace files, then it should be OK (and simpler) to have a separate makelib for
        # each file in each library.
        libraries = OrderedDict()
        for sources in self.target.content.vhdl_sources:
            library = sources.library

            if library not in libraries:
                libraries[library] = []

            files = sources.files
            libraries[library] += files

        for library, sources in libraries.items():
            if library is not None:
                cmd += ['-makelib', library]
            else:
                cmd += ['-makelib']
            cmd += sources
            cmd += ['-endlib']

        # add any additional user-supplied flags
        for flag in self.flags:
            cmd.append(str(flag))

        # write TCL file
        with open(self.cfg.xcelium_config.tcl_input_path, 'w') as f:
            f.write('run\n')
            f.write('exit\n')

        # run xrun
        print(cmd)
        call(cmd, cwd=self.cfg.build_root)

    def prepare(self):
        """ Preparation of ifxxcelium Camino wrapper script"""
        if "ifxxcelium" in self.cfg.xcelium_config.xrun:
            cmd = []
            cmd += ['ifxxcelium', 'prepare', '-uc', 'rtl']
            if self.unit:
                cmd += ["-unit", self.unit ]
            if self.id:
                cmd += ["-id", self.id ]
            else:
                self.id = "xcelium"

            print(cmd)
            call(cmd, cwd=self.cfg.build_root)

            makefile = os.environ["WORKAREA"] + "/units/" + self.unit + "/simulation/" + self.id + "/Makefile"
            if "PHONY: inicio" not in open(makefile, 'r').read():
                self.patch_makefile(makefile)
            else:
                print("inicio make target already in Makefile, will not patch it")

        else:
            print("No ifxxcelium script detected, nothing to prepare..")

    def patch_makefile(self, file):
        """ patch content of generated Makefile and append inicio target
        :type file: str
        """

        inicio_target = open( os.path.dirname(__file__) + "/xcelium.make_target", 'r')

        print(f"Patching content of Makefile: {file}")
        with open(file, 'a+') as f:
            f.write(inicio_target.read())
            f.close()



import os
import pathlib

from anasymod.util import call, back2fwd
from anasymod.generators.codegen import CodeGenerator
from anasymod.sources import VerilogSource, MEMFile, BDFile, IPRepo, EDIFFile
from anasymod.targets import FPGATarget
from anasymod.enums import FPGASimCtrl

class VivadoTCLGenerator(CodeGenerator):
    """
    Class for generating a TCL control file to setup Vivado Projects, launch Vivado and execute the generated TCL file.
    """

    def __init__(self, target: FPGATarget):
        super().__init__()
        self.target = target
        self.subst = None
        self.old_subst = None

    def create_project(self, project_name, project_directory, force=False, full_part_name=None,
                       board_part=None):
        # create the project
        cmd = ['create_project']
        if force:
            cmd.append('-force')
        cmd.append('"'+project_name+'"')
        cmd.append('"'+back2fwd(project_directory)+'"')
        if full_part_name is not None:
            cmd.extend(['-part', full_part_name])
        self.writeln(' '.join(cmd))

        # specify the board part if known
        if board_part is not None:
            if self.target.prj_cfg.board.fpga_sim_ctrl is not FPGASimCtrl.UART_ZYNQ:
                self.writeln(f'if {{[catch {{set_property board_part {board_part} [current_project]}}]}} '
                         f'{{puts "WARNING: Could not not set board_part:{board_part}"}}')
            else:
                self.writeln(f'set_property board_part {board_part} [current_project]')

    def add_project_sources(self, content):
        """
        Add sources to project.
        :param content: List of dicts that store all target specific source and define objects.
        """

        # add verilog sources
        self.add_verilog_sources(ver_src=content.verilog_sources)

        # add verilog headers
        self.add_verilog_headers(ver_hdr=content.verilog_headers)

        # add functional model HDL sources
        self.add_functional_models(fm_src=content.functional_models)

        # add vhdl sources
        self.add_vhdl_sources(vhdl_src=content.vhdl_sources)

        # add edif file
        self.add_edif_file(edif_files=content.edif_files)

        # add mem file
        self.add_mem_file(mem_files=content.mem_files)

        # add ip repo
        self.add_ip_repo(ip_repos=content.ip_repos)

        # add bd file
        self.add_bd_file(bd_files=content.bd_files)

    def add_verilog_sources(self, ver_src: [VerilogSource]):
        for src in ver_src:
            if src.files:
                self.add_files(src.files)
                if src.version is not None:
                    file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in src.files) + ' }'
                    self.set_property('file_type', value=f'{{{src.version}}}', objects=f'[get_files {file_list}]')

    def add_verilog_headers(self, ver_hdr):
        for src in ver_hdr:
            if src.files:
                self.add_files(src.files)
                file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in src.files) + ' }'
                self.set_property('file_type', '{Verilog Header}', f'[get_files {file_list}]')

    def add_functional_models(self, fm_src):
        for src in fm_src:
            if src.gen_files:
                files = []
                for file in src.gen_files:
                    files.append(file)
                self.add_files(files)

    def add_vhdl_sources(self, vhdl_src):
        for src in vhdl_src:
            if src.files:
                self.add_files(src.files)
                if src.library is not None:
                    file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in src.files) + ' }'
                    self.set_property('library', value=f'{{{src.library}}}', objects=f'[get_files {file_list}]')
                if src.version is not None:
                    file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in src.files) + ' }'
                    self.set_property('file_type', value=f'{{{src.version}}}', objects=f'[get_files {file_list}]')

    def add_edif_file(self, edif_files: [EDIFFile]):
        for edif_file in edif_files:
            if edif_file.files:
                self.add_files(edif_file.files)

    def add_mem_file(self, mem_files: [MEMFile]):
        for mem_file in mem_files:
            if mem_file.files:
                self.add_files(mem_file.files)

    def add_bd_file(self, bd_files: [BDFile]):
        for bd_file in bd_files:
            if bd_file.files:
                cmd = ['read_bd']
                cmd.append('{ ' + ' '.join('"' + back2fwd(file) + '"' for file in bd_file.files) + ' }')
                self.writeln(' '.join(cmd))

    def add_ip_repo(self, ip_repos: [IPRepo]):
        for ip_repo in ip_repos:
            if ip_repo.files:
                self.set_property(name='ip_repo_paths', value='{ ' + ' '.join('"' + back2fwd(file) + '"' for file in ip_repo.files) + ' }', objects='[current_project]')
        self.writeln('update_ip_catalog -rebuild')

    def add_project_defines(self, content, fileset):
        """
        Add defines to project.
        :param content: List of dicts that store all target specific source and define objects.
        """
        define_list = []
        for define in content.defines:
            for k, v in define.define.items():
                if v is not None:
                    define_list.append(f"{k}={v}")
                else:
                    define_list.append(f"{k}")

        self.set_property('verilog_define', f"{{{' '.join(define_list)}}}", fileset)

    def add_files(self, files, norecurse=True, fileset=None):
        if files is None or len(files) == 0:
            # don't generate a command because add_files does not work with
            # an empty list of files
            return

        cmd = ['add_files']
        if fileset is not None:
            cmd.extend(['-fileset', fileset])
        if norecurse:
            cmd.append('-norecurse')
        cmd.append('{ '+' '.join('"'+back2fwd(file)+'"' for file in files)+' }')
        self.writeln(' '.join(cmd))

    def add_include_dirs(self, content, objects):
        # otherwise, generate list of all include directories
        inc_dirs = []
        for include_dir in content.include_dirs:
            inc_dirs += include_dir.files

        # do nothing if there are no include directories
        if len(inc_dirs) == 0:
            return

        # format into a TCL argument
        inc_dirs = [back2fwd(inc_dir) for inc_dir in inc_dirs]  # fix backslashes
        inc_dirs = ['"' + str(inc_dir) + '"' for inc_dir in inc_dirs]  # add quotes
        inc_dirs = ' '.join(inc_dirs)  # separate by spaces
        inc_dirs = '{ ' + inc_dirs + ' }'  # surround with curly braces

        # then set the include_dirs property
        self.set_property(name='include_dirs', value=inc_dirs, objects=objects)

    def set_property(self, name, value, objects):
        self.writeln(' '.join(['set_property', '-name', name, '-value', value, '-objects', objects]))

    def subst_path(self, drive):
        self.subst = drive
        if os.path.exists(self.subst):
            print(self.subst, " already exists, saving copy of existing subst path")
            self.old_subst = back2fwd(str(pathlib.Path(self.subst).resolve()))
        # deleting old subst
        self.writeln('[catch {exec subst ' + self.subst + ' /d}]')
        self.writeln('exec subst ' + self.subst + ' ' + back2fwd(os.path.dirname(self.target.project_root)))
        project_root = self.subst + r"\\" + self.target.prj_cfg.vivado_config.project_name
        return project_root

    def run(self, filename=r"run.tcl", nolog=True, nojournal=True, stack=None,
            interactive=False, err_str=None, return_error=False):
        # write the TCL script
        tcl_script = os.path.join(self.target.prj_cfg.build_root, filename)
        self.write_to_file(tcl_script)

        # assemble the command
        cmd = []
        cmd += [self.target.prj_cfg.vivado_config.vivado]

        # Specify stack size if needed.  Xilinx suggests this
        # can sometimes fix segfault issues:
        # https://www.xilinx.com/support/answers/64434.html
        if stack:
            cmd += ['-stack', f'{stack}']

        # run mode
        if interactive:
            cmd += ['-mode', 'tcl']
        else:
            cmd += ['-mode', 'batch']

        # script location
        cmd += ['-source', tcl_script]

        # inserting lsf_opts after vivado call:
        cmd[1:1] = self.target.prj_cfg.vivado_config.lsf_opts.split()

        if nolog:
            cmd.append('-nolog')
        if nojournal:
            cmd.append('-nojournal')

        # run the script
        return call(args=cmd, cwd=self.target.prj_cfg.build_root, err_str=err_str, return_error=return_error)

    @property
    def version_year(self):
        return self.target.prj_cfg.vivado_config.version_year

    @property
    def version_number(self):
        return self.target.prj_cfg.vivado_config.version_number


import os, sys
import time

from anasymod.config import EmuConfig
from anasymod.structures.structure_config import StructureConfig
from .console_print import cprint_block_start, cprint_block_end

class CtrlApi:
    """
    Start an interactive control interface to HW target for running regression tests or design exploration/debug.
    For FPGA/Emulators, as a pre-requisit, bitstream must have been created and programmed. Additionally any eSW
    necessary in the targeted system must have already been programmed.
    """
    def __init__(self, cwd, pcfg: EmuConfig, scfg: StructureConfig, prompt, debug):
        self.pcfg=pcfg
        self.scfg=scfg
        self.cwd = cwd
        self.prompt = prompt
        self.debug = debug

    ### User Functions

    def sendline(self, line, timeout=float('inf')):
        """
        Send a single line in target shell specific language e.g. in tcl for tcl shell.
        :param line: Line that shall be send to/processed by shell
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def source(self, script, timeout=float('inf')):
        """
        Source a script written language for targeted shell.
        :param script: Name/Path to script that shall be sourced
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def setup_trace_unit(self, trigger_name, trigger_operator, trigger_value, sample_decimation=None, sample_count=None):
        """
        Setup the trace unit. This involves defining the signal, that shall be used to start tracing and defining the
        comparison operator, which is used to monitor the trigger signal.

        Note: In case the trigger value is set too high and an overflow occurs, the simulation will hang, as the trigger
        condition will never be met.

        :param trigger_name:        Probe signal to be used as trigger signal.
        :param trigger_operator:    Comparison operator that is used when monitoring the trigger signal to detect,
                                    when a trigger shall be fired. Available oerators are:
                                        eq (equal),
                                        neq (not equal),
                                        gt (greater than),
                                        gteq (greater or equal than),
                                        lt (lesser than),
                                        lteq (lesser or equal than)
        :param trigger_value:       Value, the probe signal is compared to.
        :param sample_decimation:   Number of samples to be skipped during recording. By default, every sample will
                                    stored in the results file, adding a sample_decimation will allow to only record
                                    every x sample.
        :param sample_count:        Number of samples to be recorded. This number shall not exceed that maximum ILA
                                    depth defined during project setup.
        """
        raise NotImplementedError("Base class was called to execute function")

    def arm_trace_unit(self):
        """
        Arm the trace unit, this will delete the buffer and arm the trigger.
        """
        raise NotImplementedError("Base class was called to execute function")

    def wait_on_and_dump_trace(self, result_file=None):
        """
        Wait until the trace unit stopped recording data. Transmit this data to the host PC, store by default to the raw
        result file path, or a custom path provided by the user, and convert analog values from fixed-point to float.
        Finally store it to a .vcd file in the default location.

        :param result_file: Optionally, it is possible to provide a custom result file path.
        """
        raise NotImplementedError("Base class was called to execute function")

    def refresh_param(self, name, timeout=30):
        """
        Refresh selected control parameter.
        :param name: Name of control parameter
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def get_param(self, name, timeout=30):
        """
        Read value of a control parameter in design.
        :param name: Name of control parameter to be read
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def set_param(self, name, value, timeout=30):
        """
        Set value of a control parameter in design.
        :param name: Name of control parameter to be set
        :param value: Value of control parameter sto be set
        :param timeout: Maximum time granted for operation to finish
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def set_var(self, name, value):
        """
        Define a variable in target shell environment.
        :param name: Name of variable that shall be set
        :param value: Value of variable that shall be set
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def set_reset(self, value, timeout=30):
        """
        Control the 'emu_rst' signal, in order to put the system running on the FPGA into or out of reset state.
        :param value: Value of reset signal, 1 will set it to reset and 0 will release reset.
        :param timeout: Maximum time granted for operation to finish
        """
        raise NotImplementedError("Base class was called to execute function")

    def get_emu_time_int(self, timeout=30):
        """
        Get current time of the FPGA simulation as an unscaled integer value.
        :param timeout: Maximum time granted for operation to finish
        """
        raise NotImplementedError("Base class was called to execute function")

    def get_emu_time(self, timeout=30):
        """
        Get current time of the FPGA simulation as a decimal value.
        :param timeout: Maximum time granted for operation to finish
        """
        return self.get_emu_time_int(timeout=timeout) * self.pcfg.cfg.dt_scale

    def set_ctrl_mode(self, value, timeout=30):
        """
        Set the control mode that shall be applied to stall the FPGA simulation.
        :param value: Integer value setting the currently active control mode:
                        0:  FPGA simulation is not stalled
                        1:  FPGA simulation is immediately stalled
                        2:  FPGA simulation is stalled, after time stored in *ctrl_data* has passed, starting from
                            the point in time this mode has been selected.
                        3:  FPGA simulation is stalled, once time value stored in *ctrl_data* has been reached
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_param(name=self.scfg.emu_ctrl_mode.name, value=value, timeout=timeout)

    def set_ctrl_data(self, value, timeout=30):
        """
        Set a time value as unscaled integer.
        :param value:  Unscaled integer value that represents a time. This value is interpreted according to
        the selected ctrl_mode
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_param(name=self.scfg.emu_ctrl_data.name, value=value, timeout=timeout)

    def stall_emu(self, timeout=30):
        """
        Stall the FPGA simulation immediately.
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_ctrl_mode(1, timeout=timeout)

    def sleep_emu(self, t, timeout=30):
        """
        Stall FPGA simulation, after emulated time of *t* has passed, starting from the point in time this function was
        called.
        :param t: Time value that shall pass before FPGA simulation is stalled.
        :param timeout: Maximum time granted for operation to finish
        """

        # stall
        self.stall_emu()

        # set up in sleep mode
        t_next = t + self.get_emu_time(timeout=timeout)
        t_next_int = int(round(t_next / self.pcfg.cfg.dt_scale))
        self.set_ctrl_data(t_next_int)
        self.set_ctrl_mode(2)

        # wait for enough time to pass
        while(self.get_emu_time_int() < t_next_int):
            pass

    ### Utility Functions

    def _initialize(self):
        """
        Initialize the control interface, this is usually done after the bitstream was programmed successfully on the FPGA.
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def _setup_ctrl(self, server_addr):
        """
        Prepare instrumentation on the FPGA to allow interactive control.
        :param server_addr: Address of remote hardware server
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")

    def _expect_prompt(self, timeout=float('inf')):
        """
        Wait for the shell used to transmit commands to provide a response.
        :param timeout: Maximum time granted for shell to open
        :return:
        """
        before = ''
        start_time = time.time()
        lines_recv = 0
        while (time.time() - start_time) < timeout:
            remaining = timeout - (time.time() - start_time)
            if remaining == float('inf'):
                remaining = None
            index = self.proc.expect(['\n', self.prompt], timeout=remaining)
            if index == 0:
                before += self.proc.before + '\n'
                lines_recv += 1
                if self.debug:
                    if lines_recv == 2:
                        cprint_block_start('RECV', 'cyan')
                    if lines_recv >= 2:
                        print(self.proc.before)
            else:
                break
        if self.debug and lines_recv >= 2:
            cprint_block_end('RECV', 'cyan')
        return before

    def _initialize_vivado_tcl(self):
        """
        Initialize the control interface, this is usually done after the bitstream was programmed successfully on the FPGA.
        :return:
        """

        # log current status
        print('Starting Vivado TCL interpreter.')
        sys.stdout.flush()

        # construct the command to launch Vivado
        cmd = 'vivado '
        cmd += self.pcfg.vivado_config.lsf_opts_ls + ' '
        cmd += '-nolog -nojournal -notrace -mode tcl'

        # Use pexpect under linux for interactive vivado ctrl
        if os.name == 'posix':
            # Add Vivado to the path using the Windows PATH separator (semicolon)
            # A copy of the environment is made to avoid side effects outside this function
            # TODO: do this only if needed
            env = os.environ.copy()
            env['PATH'] += f':{os.path.dirname(self.pcfg.vivado_config.vivado)}'
            # Launch Vivado
            from pexpect import spawnu
            self.proc = spawnu(command=cmd, cwd=self.cwd, env=env)
        elif os.name == 'nt':
            # Add Vivado to the path using the Windows PATH separator (semicolon)
            # A copy of the environment is made to avoid side effects outside this function
            # TODO: do this only if needed
            env = os.environ.copy()
            env['PATH'] += f';{os.path.dirname(self.pcfg.vivado_config.vivado)}'
            os.environ['WEXPECT_SPAWN_CLASS'] = 'SpawnPipe'
            # Launch Vivado
            try:
                # import patched wexpect from Inicio installation
                from site_pip_packages.wexpect import spawn
            except:
                from wexpect import spawn
            self.proc = spawn(command=cmd, cwd=self.cwd, env=env)
        else:
            raise Exception(f'No supported OS was detected, supported OS for interactive control are windows and linux.')

        # wait for the prompt
        self._expect_prompt(timeout=300)

    def __del__(self):
        """
        Close connection to shell.
        :return:
        """
        raise NotImplementedError("Base class was called to execute function")


import os, yaml

from anasymod.enums import ConfigSections
from anasymod.base_config import BaseConfig
from anasymod.config import EmuConfig
from anasymod.sim_ctrl.datatypes import (
    DigitalSignal, DigitalCtrlInput, DigitalCtrlOutput,
    AnalogCtrlInput, AnalogCtrlOutput, AnalogProbe)

class StructureConfig():
    """
    In this configuration, all the toplevel information about the generated toplevel is included.
    It will be used for generation of the target specific top-level, as well as attached wrappers.

    There is also a specific interface to flow plugins that allows modification due to some needs
    from the plugin side, e.g. additional clks, resets, ios to the host application or resources on the FPGA board.
    """
    def __init__(self, prj_cfg: EmuConfig, simctrl_path, can_use_default_oscillator=True):
        # Internal variables
        self.i_addr_counter = 0
        self.o_addr_counter = 0

        # Path to clks.yaml file
        self._clks_file_path = os.path.join(prj_cfg.root, 'clks.yaml')
        # Path to simctrl.yaml file
        self._simctrl_file_path = simctrl_path

        self.cfg = Config(prj_cfg=prj_cfg)
        self.cfg.update_config()

        #########################################################
        # Manage clks
        #########################################################

        self.emu_clk = ClkIndependent(name='emu_clk', freq=float(prj_cfg.cfg.emu_clk_freq))
        self.emu_clk_2x = ClkIndependent(name='emu_clk_2x', freq=float(prj_cfg.cfg.emu_clk_freq) * 2) # multiplied by two, as emu_clk_2x is twice as fast as emu_clk
        self.dbg_clk = ClkIndependent(name='dbg_hub_clk', freq=float(prj_cfg.board.dbg_hub_clk_freq))

        # add clk_in
        self.clk_i_num = len(prj_cfg.board.clk_pin)

        # clk_in names cannot be changed
        if self.clk_i_num == 2:
            self.clk_i = [DigitalSignal(abspath=None, width=1, name='clk_in1_p'), DigitalSignal(abspath=None, width=1, name='clk_in1_n')]
        elif self.clk_i_num == 1:
            self.clk_i = [DigitalSignal(abspath=None, width=1, name='clk_in1')]
        else:
            raise ValueError(
                f"Wrong number of pins for boards param 'clk_pin', expecting 1 or 2, provided:{self.clk_i_num}")

        # Add independent clks, that will be implemented via clk_wiz IP core for FPGA
        self.clk_independent = []
        """ type : [ClkIndependent]"""

        # Add derived clks; those are aligned with emu_clk
        self.clk_derived = []
        """ type : [ClkDerived]"""

        # Number of gated clks
        self.num_gated_clks = 0

        # Number of dt requests
        self.num_dt_reqs = 0

        self._read_clksfile()

        # add a default derived clock if needed
        if (self.num_dt_reqs == 0) and can_use_default_oscillator:
            self.use_default_oscillator = True
        else:
            self.use_default_oscillator = False

        # make note of whether this is an UltraScale board
        self.is_ultrascale = prj_cfg.board.is_ultrascale

        #########################################################
        # Simulation control interfaces
        #########################################################

        # CtrlIOs
        self.digital_ctrl_inputs = []
        self.digital_ctrl_outputs = []
        self.analog_ctrl_inputs = []
        self.analog_ctrl_outputs = []
        self.analog_probes = []
        self.digital_probes = []

        # annotate special control signals that should not
        # be routed to sim_ctrl
        self.special_ctrl_ios = set()

        # Add time signal representing current simulated time
        self.time_probe = DigitalSignal(
            name='emu_time',
            width=prj_cfg.cfg.time_width,
            abspath=''
        )

        # Add DigitalCtrlInput for reset
        self.reset_ctrl = DigitalCtrlInput(
            abspath=None,
            name='emu_rst',
            width=1)
        self.reset_ctrl.i_addr = self._assign_i_addr()
        self.digital_ctrl_inputs += [self.reset_ctrl]
        self.special_ctrl_ios.add(self.reset_ctrl.name)

        # Add DigitalCtrlInput for control signal 'emu_dec_thr' to manage decimation
        # ratio for capturing probe samples
        self.dec_thr_ctrl = DigitalCtrlInput(
            abspath=None,
            name='emu_dec_thr',
            width=int(prj_cfg.cfg.dec_bits)
        )
        self.dec_thr_ctrl.i_addr = self._assign_i_addr()
        self.digital_ctrl_inputs += [self.dec_thr_ctrl]
        self.special_ctrl_ios.add(self.dec_thr_ctrl.name)

        # Add DigitalCtrlInput for control signal 'emu_time_tgt' to run for
        # a specific amount of time
        self.emu_ctrl_data = DigitalCtrlInput(
            name='emu_ctrl_data',
            width=int(prj_cfg.cfg.time_width),
            abspath = None
        )
        self.emu_ctrl_data.i_addr = self._assign_i_addr()
        self.digital_ctrl_inputs += [self.emu_ctrl_data]
        self.special_ctrl_ios.add(self.emu_ctrl_data.name)

        # Add DigitalCtrlInput for control signal 'emu_ctrl_mode' to run for
        # a specific amount of time
        self.emu_ctrl_mode = DigitalCtrlInput(
            name='emu_ctrl_mode',
            width=2,
            abspath = None
        )
        self.emu_ctrl_mode.i_addr = self._assign_i_addr()
        self.digital_ctrl_inputs += [self.emu_ctrl_mode]
        self.special_ctrl_ios.add(self.emu_ctrl_mode.name)

        # Add DigitalCtrlOutput for reading the emulation time
        self.emu_time_vio = DigitalCtrlOutput(
            name='emu_time_vio',
            width=prj_cfg.cfg.time_width,
            abspath = 'emu_time'
        )
        self.emu_time_vio.i_addr = self._assign_o_addr()
        self.digital_ctrl_outputs += [self.emu_time_vio]
        self.special_ctrl_ios.add(self.emu_time_vio.name)

        # Add DigitalSignal for control of signal 'emu_dec_cmp' to trigger sampling
        # for the ila depending on 'emu_dec_thr'
        self.dec_cmp = DigitalSignal(
            name='emu_dec_cmp',
            abspath='emu_dec_cmp_probe',
            width=1
        )

        self._read_simctrlfile()

    def _assign_i_addr(self):
        """
        Function to assign an input address to an Input object.
        """
        curr_addr = self.i_addr_counter
        self.i_addr_counter +=1
        return curr_addr

    def _assign_o_addr(self):
        """
        Function to assign an input address to an Input object.
        """
        curr_addr = self.o_addr_counter
        self.o_addr_counter +=1
        return curr_addr

    def _read_clksfile(self):
        """
        Read all lines from clks file clks.config and store in structure config attributes. This includes either
        independent clks, which will directly be added to the clk management block and in case of a FPGA simulation
        implemented via clk_wiz IP core, or derived clks, which will be derived for the main emulation clk em,u_clk_2x.
        """
        if os.path.isfile(self._clks_file_path):
            try:
                clks = yaml.safe_load(open(self._clks_file_path, "r"))
            except yaml.YAMLError as exc:
                raise Exception(exc)

            if clks is not None:
                # Add independent clks to structure config
                if 'independent_clks' in clks.keys():
                    print(f'Independent Clks: {[key for key in clks["independent_clks"].keys()]}')
                    for independent_clk in clks['independent_clks'].keys():
                        self.clk_independent.append(
                            ClkIndependent(
                                name=independent_clk,
                                freq=float(clks['independent_clks'][independent_clk]['freq'])
                            )
                        )
                else:
                    print(f'No Independent Clks provided.')

                # Add derived clks to structure config
                if 'derived_clks' in clks.keys():
                    print(f'Derived Clks: {[key for key in clks["derived_clks"].keys()]}')

                    for derived_clk_name, derived_clk in clks['derived_clks'].items():
                        # find out the absolute path prefix
                        if 'abspath' in derived_clk:
                            abspath_default = derived_clk['abspath']
                        else:
                            raise Exception(f'No abspath provided for clk: {derived_clk}')

                        # little convenience function for prepending the absolute path
                        def make_abs_path(key, default):
                            signame = derived_clk.get(key, '')
                            if signame == '':
                                signame = default
                            return f'{abspath_default}.{signame}'

                        # read out the prefix name (if any)
                        preset = derived_clk.get('preset', None)

                        # for each subsequent signal, set its absolute path
                        # if provided explicitly or implicitly via a preset

                        if ('emu_dt' in derived_clk) or \
                                (preset in {'variable_timestep', 'oscillator'}):
                            abspath_emu_dt = make_abs_path('emu_dt', '__emu_dt')
                        else:
                            abspath_emu_dt = None

                        if ('dt_req' in derived_clk) or \
                                (preset in {'variable_timestep', 'oscillator'}):
                            abspath_dt_req = make_abs_path('dt_req', '__emu_dt_req')
                            self.num_dt_reqs += 1
                        else:
                            abspath_dt_req = None

                        if ('emu_clk' in derived_clk) or \
                                (preset in {'fixed_timestep', 'variable_timestep', 'oscillator'}):
                            abspath_emu_clk = make_abs_path('emu_clk', '__emu_clk')
                        else:
                            abspath_emu_clk = None

                        if ('emu_rst' in derived_clk) or \
                                (preset in {'fixed_timestep', 'variable_timestep', 'oscillator'}):
                            abspath_emu_rst = make_abs_path('emu_rst', '__emu_rst')
                        else:
                            abspath_emu_rst = None

                        if ('gated_clk' in derived_clk) or \
                                (preset in {'fixed_timestep', 'oscillator'}):
                            abspath_gated_clk = make_abs_path('gated_clk', '__emu_clk_i')
                            self.num_gated_clks += 1
                        else:
                            abspath_gated_clk = None

                        if ('gated_clk_req' in derived_clk) or \
                                (preset in {'fixed_timestep', 'oscillator'}):
                            abspath_gated_clk_req = make_abs_path('gated_clk_req', '__emu_clk_val')
                        else:
                            abspath_gated_clk_req = None

                        self.clk_derived.append(
                            ClkDerived(
                                name=derived_clk_name,
                                abspath_emu_dt=abspath_emu_dt,
                                abspath_emu_clk=abspath_emu_clk,
                                abspath_emu_rst=abspath_emu_rst,
                                abspath_dt_req=abspath_dt_req,
                                abspath_gated_clk=abspath_gated_clk,
                                abspath_gated_clk_req=abspath_gated_clk_req
                            )
                        )
                else:
                    print(f'No Derived Clks provided.')
        else:
            print(f"No clks.config file existing, no additional clks will be added for this design.")

    def _read_simctrlfile(self):
        """
        Read all lines from simulation control file simctrl.yaml and store in structure config attributes.
        """

        if os.path.isfile(self._simctrl_file_path):
            try:
                sigs = yaml.safe_load(open(self._simctrl_file_path, "r"))
            except yaml.YAMLError as exc:
                raise Exception(exc)
        else:
            print('No simctrl.yaml file existing, no additional probes will be available for this simulation.')
            return

        if sigs is None:
            print('No signals specified in simctrl.yaml.')
            return
        # Add analog probes to structure config
        if 'analog_probes' in sigs:
            print(f'Analog Probes: {list(sigs["analog_probes"].keys())}')
            for name, analog_probe in sigs['analog_probes'].items():
                self.analog_probes.append(AnalogProbe.from_dict(name, analog_probe))
            else:
                print(f'No Analog Probes provided.')

        # Add digital probes to structure config
        if 'digital_probes' in sigs:
            print(f'Digital Probes: {list(sigs["digital_probes"].keys())}')
            for name, digital_probe in sigs['digital_probes'].items():
                self.digital_probes.append(DigitalSignal.from_dict(name, digital_probe))
        else:
            print(f'No Digital Probes provided.')

        # Add digital ctrl inputs to structure config
        if sigs.get('digital_ctrl_inputs', None) is not None:
            print(f'Digital Ctrl Inputs: {list(sigs["digital_ctrl_inputs"].keys())}')
            for name, d_ctrl_in in sigs['digital_ctrl_inputs'].items():
                d_ctrl_i = DigitalCtrlInput.from_dict(name, d_ctrl_in)
                d_ctrl_i.i_addr = self._assign_i_addr()
                self.digital_ctrl_inputs.append(d_ctrl_i)
        else:
            print(f'No Digital Ctrl Input provided.')

        # Add digital ctrl outputs to structure config
        if sigs.get('digital_ctrl_outputs', None) is not None:
            print(f'Digital Ctrl Outputs: {list(sigs["digital_ctrl_outputs"].keys())}')
            for name, d_ctrl_out in sigs['digital_ctrl_outputs'].items():
                d_ctrl_o = DigitalCtrlOutput.from_dict(name, d_ctrl_out)
                d_ctrl_o.o_addr = self._assign_o_addr()
                self.digital_ctrl_outputs.append(d_ctrl_o)
        else:
            print(f'No Digital Ctrl Outputs provided.')

        # Add analog ctrl inputs to structure config
        if sigs.get('analog_ctrl_inputs', None) is not None:
            print(f'Analog Ctrl Inputs: {list(sigs["analog_ctrl_inputs"].keys())}')
            for name, a_ctrl_in in sigs['analog_ctrl_inputs'].items():
                a_ctrl_i = AnalogCtrlInput.from_dict(name, a_ctrl_in)
                a_ctrl_i.i_addr = self._assign_i_addr()
                self.analog_ctrl_inputs.append(a_ctrl_i)
        else:
            print(f'No Analog Ctrl Input provided.')

        # Add analog ctrl outputs to structure config
        if sigs.get('analog_ctrl_outputs', None) is not None:
            print(f'Analog Ctrl Outputs: {list(sigs["analog_ctrl_outputs"].keys())}')
            for name, a_ctrl_out in sigs['analog_ctrl_outputs'].items():
                a_ctrl_o = AnalogCtrlOutput.from_dict(name, a_ctrl_out)
                a_ctrl_o.o_addr = self._assign_o_addr()
                self.analog_ctrl_outputs.append(a_ctrl_o)
        else:
            print(f'No Analog Ctrl Outputs provided.')

class ClkIndependent(DigitalSignal):
    """
    Container for an independent clk object.
    """

    def __init__(self, name, freq):
        # call the super constructor
        super().__init__(abspath='', name=name, width=1)

        # save settings
        self.freq = freq


class ClkDerived(DigitalSignal):
    """
    Container for a derived clk object.
    """

    def __init__(self, name, abspath_emu_dt=None, abspath_emu_clk=None,
                 abspath_emu_rst=None, abspath_dt_req=None,
                 abspath_gated_clk=None, abspath_gated_clk_req=None):
        # call the super constructor
        super().__init__(abspath='', name=name, width=1)

        # save settings
        self.abspath_emu_dt = abspath_emu_dt
        self.abspath_emu_clk = abspath_emu_clk
        self.abspath_emu_rst = abspath_emu_rst
        self.abspath_dt_req = abspath_dt_req
        self.abspath_gated_clk = abspath_gated_clk
        self.abspath_gated_clk_req = abspath_gated_clk_req


class Config(BaseConfig):
    """
    Container to store all config attributes.
    """

    def __init__(self, prj_cfg: EmuConfig):
        super().__init__(cfg_file=prj_cfg.cfg_file, section=ConfigSections.STRUCTURE)

        self.rst_clkcycles = 0
        """ type(int) : number of clk cycles, the initial emu_rst signal shall be active. """


import os

from anasymod.viewer.viewer import Viewer
from anasymod.util import call

class SimVisionViewer(Viewer):
    def view(self):
        # build command
        if os.path.isfile(self.target.cfg.vcd_path):
            cmd = [self.cfg.simvision_config.simvision, '-wave', self.target.cfg.vcd_path]
            cmd[1:1] = self.cfg.simvision_config.lsf_opts.split()
            # add waveform file if it exists
            if self.cfg.simvision_config.svcf_config is not None:
                if os.path.isfile(self.cfg.simvision_config.svcf_config):
                    cmd.extend(['-input', self.cfg.simvision_config.svcf_config])

            # run command
            call(cmd)
        else:
            raise Exception(f'ERROR: Result file: {self.target.cfg.vcd_path} does not exist; cannot open waveforms!')


import shutil, zipfile, filecmp, errno, time
import os.path
import yaml
import numpy as np

from argparse import ArgumentParser

from anasymod.config import EmuConfig, SimVisionConfig, XceliumConfig
from anasymod.sim.vivado import VivadoSimulator
from anasymod.sim.icarus import IcarusSimulator
from anasymod.sim.xcelium import XceliumSimulator
from anasymod.viewer.gtkwave import GtkWaveViewer
from anasymod.viewer.scansion import ScansionViewer
from anasymod.viewer.simvision import SimVisionViewer
from anasymod.emu.vivado_emu import VivadoEmulation
from anasymod.emu.xsct_emu import XSCTEmulation
from anasymod.files import get_full_path, get_from_anasymod, mkdir_p
from anasymod.sources import *
from anasymod.filesets import Filesets
from anasymod.defines import Define
from anasymod.targets import CPUTarget, FPGATarget
from anasymod.enums import ConfigSections, FPGASimCtrl
from anasymod.utils import statpro
from anasymod.util import expand_path
from anasymod.wave import ConvertWaveform
from anasymod.plugins import Plugin
from typing import Union
from importlib import import_module

class Analysis():
    """
    This is the top user Class that shall be used to exercise anasymod.
    """
    def __init__(self, input=None, build_root=None, simulator_name=None,
                 synthesizer_name=None, viewer_name=None, preprocess_only=None,
                 op_mode=None, active_target=None):

        # Parse command line arguments
        self.args = None
        self._parse_args()

        # Initialize attributes
        self.float_type = False # Defines which data type is used for functional models; default is fixed-point
        self._plugin_args = {} # Namespace object including all options set for generators

        # Overwrite input location in case it was provided when instantiation the Analysis class
        if input is not None:
            self.args.input = input

        # expand path of input and output directories relative to analysis.py
        self.args.input = get_full_path(self.args.input)

        # update args according to user specified values when instantiating analysis class
        self.args.simulator_name = simulator_name if simulator_name is not None else self.args.simulator_name
        self.args.synthesizer_name = synthesizer_name if synthesizer_name is not None else self.args.synthesizer_name
        self.args.viewer_name = viewer_name if viewer_name is not None else self.args.viewer_name
        self.args.preprocess_only = preprocess_only if preprocess_only is not None else self.args.preprocess_only
        self.args.active_target = active_target if active_target is not None else self.args.active_target

        # Load config file
        self.cfgfile_path = os.path.join(self.args.input, 'prj.yaml')

        if os.path.isfile(self.cfgfile_path):
            try:
                self.cfg_file = yaml.safe_load(open(self.cfgfile_path, "r"))
            except yaml.YAMLError as exc:
                raise Exception(exc)
        else:
            self.cfg_file = None
            print(f"Warning: no config file was found for the project, expected path is: {self.cfgfile_path}")

        # Initialize Targets
        self.act_fpga_target = 'fpga'
        self.fpga_targets = [self.act_fpga_target]
        try:
            for custom_target in self.cfg_file[ConfigSections.FPGA_TARGET].keys():
                if custom_target not in self.fpga_targets:
                    self.fpga_targets.append(custom_target)
        except:
            pass

        self.act_cpu_target = 'sim'
        self.cpu_targets = [self.act_cpu_target]
        try:
            for custom_target in self.cfg_file[ConfigSections.CPU_TARGET].keys():
                if custom_target not in self.cpu_targets:
                    self.cpu_targets.append(custom_target)
        except:
            pass

        # Initialize dict for tracking, which targets are already setup.
        self._setup_finished = {}
        for target in self.cpu_targets + self.fpga_targets:
            self._setup_finished[target] = False

        self.project_sources_finalized = False

        # Initialize project config
        self._prj_cfg = EmuConfig(root=self.args.input, cfg_file=self.cfg_file, active_target=self.args.active_target, build_root=build_root)

        # Initialize Plugins
        self._plugins = []
        for plugin in self._prj_cfg.cfg.plugins:
            try:
                i = import_module(f"{plugin}.plugin")
                inst = i.CustomPlugin(prj_cfg=self._prj_cfg, cfg_file=self.cfg_file, prj_root=self.args.input)
                self._plugins.append(inst)
                setattr(self, inst._name, inst)
            except:
                raise KeyError(f"Could not process plugin:{plugin} properly! Check spelling")

        #Set active target
        self.set_target(self.args.active_target)

        # Initialize filesets, those can later be modified via the add_sources function
        self._setup_filesets()

        # Check which mode is used to run, in case of commandline mode, besides setting up the class, also argument will be processed and executed
        if op_mode in ['commandline']:
            print(f"Running in commandline mode.")


            ###############################################################
            # Process command line arguments for plugins
            ###############################################################

            for plugin in self._plugins:
                # Parse command line arguments and execute all actions for each plugin
                plugin._parse_args()
                args = plugin._return_args()

                for arg in args.__dict__:
                    plugin.set_option(name=arg, value= args.__dict__[arg]) # Set options for the generator according to commandline arguments
                    self._plugin_args[arg] = args.__dict__[arg]

            # Set float type to true, in case floating-point data types are used during simulation.
            # This is needed when converting result files.
            if 'float' in self._plugin_args.keys():
                self.float_type = self._plugin_args['float']

            ###############################################################
            # Set options from to command line arguments
            ###############################################################

            self._prj_cfg.cfg.preprocess_only = self.args.preprocess_only

            ###############################################################
            # Execute actions according to command line arguments
            ###############################################################

            # generate source code, e.g. functional models via msdsl
            if self.args.models:
                self.gen_sources()

            # generate bitstream
            if self.args.build:
                self.set_target(self.act_fpga_target)
                self.build()

            # run FPGA if desired
            if self.args.emulate:
                self.set_target(self.act_fpga_target)
                self.emulate()

            # launch FPGA if desired
            if self.args.launch:
                self.set_target(self.act_fpga_target)
                self.launch()

            # run simulation if desired
            if self.args.sim or self.args.preprocess_only:
                self.set_target(self.act_cpu_target)
                self.simulate(unit=self.args.unit, id=self.args.id)

            # view results if desired
            if self.args.view and (self.args.sim or self.args.preprocess_only):
                self.set_target(self.act_cpu_target)
                self.view()

            if self.args.view and self.args.emulate:
                self.set_target(self.act_fpga_target)
                self.view()

##### Functions exposed for user to exercise on Analysis Object

    def add_sources(self, sources: Union[Sources, Define, list]):
        """
        Function to add sources or defines to filesets. This will also retrigger fileset dict population

        :param sources: Individual source or list of sources, that shall be added to filesets
        """

        if not isinstance(sources, list):
            sources = [sources]
        for source in sources:
            if isinstance(source, VerilogSource):
                self.filesets._verilog_sources.append(source)
            elif isinstance(source, VerilogHeader):
                self.filesets._verilog_headers.append(source)
            elif isinstance(source, VHDLSource):
                self.filesets._vhdl_sources.append(source)
            elif isinstance(source, IncludeDir):
                self.filesets._include_dirs.append(source)
            elif isinstance(source, Define):
                self.filesets._defines.append(source)
            elif isinstance(source, EDIFFile):
                self.filesets._edif_files.append(source)
            elif isinstance(source, FirmwareFile):
                self.filesets._firmware_files.append(source)
            elif isinstance(source, XCIFile):
                self.filesets._xci_files.append(source)
            elif isinstance(source, TCLFile):
                self.filesets._tcl_files.append(source)
            elif isinstance(source, XDCFile):
                self.filesets._xdc_files.append(source)
            elif isinstance(source, MEMFile):
                self.filesets._mem_files.append(source)
            elif isinstance(source, BDFile):
                self.filesets._bd_files.append(source)
            elif isinstance(source, IPRepo):
                self.filesets._ip_repos.append(source)
            elif isinstance(source, FunctionalModel):
                source.set_gen_files_path(hdl_dir_root=self._prj_cfg.build_root_functional_models)
                self.filesets._functional_models.append(source)
            else:
                print(f'WARNING: Provided source:{source} does not have a valid type, skipping this command!')

        self.project_sources_finalized = False

    def set_target(self, target_name):
        """
        Changes the domainspecific active target to target_name, e.g. the active CPU target to target_name.

        :param target_name: name of target, that shall be set as active for the respective domain.
        """
        self.args.active_target = target_name
        if self.args.active_target in self.fpga_targets:
            self.act_fpga_target = self.args.active_target
        elif self.args.active_target in self.cpu_targets:
            self.act_cpu_target = self.args.active_target
        else:
            raise Exception(f'Active target:{self.args.active_target} is not available for project, please declare the target first in the project configuration.')

        self._prj_cfg._update_build_root(active_target=target_name)

    def set_generator_option(self, generator, name, value):
        """
        Set an option for a source code generator that is added to the project, e.g. msdsl.

        :param generator:   Name of the generator for which an option shall be set. The name needs to be identical to
                            the one provided in the prj.yaml file.
        :param name:        Option name that shall be set
        :param value:       Value the option shall be set to
        """

        valid_plugin = None
        """ : type : Plugin"""

        for plugin in self._plugins:
            if generator == plugin._name:
                valid_plugin = plugin

        # Check if any valid plugin was selected
        if not valid_plugin:
            raise Exception(f'ERROR: Provided plugin/s:{generator} were not registered in the project configuration!')

        # Call the plugin set_option function to propagate the option change
        valid_plugin.set_option(name=name, value=value)

        # Store the option change in analysis object
        self._plugin_args[name] = value

        # In case the float option is set, also update the float_type instance attribute
        # ToDo: A more generic approach how to deal with this kind of change might be needed -> querry self.plugin_args dict instead
        if name == 'float':
            self.float_type = value

    def gen_sources(self, plugins=None):
        """
        Run all plugin generators added to the project ro generate source code. If parameter plugin is set to None,
        all generators will be run. Otherwise only the list of plugins provided will be run.

        :param plugin: List of plugin generator names that shall be run.
        """

        valid_plugins = []
        if plugins is None:
            valid_plugins = self._plugins
        elif isinstance(plugins, str):
            for prj_plugin in self._plugins:
                if plugins == prj_plugin._name:
                    valid_plugins = [prj_plugin]

        elif isinstance(plugins, list):
            for prj_plugin in self._plugins:
                for plugin in plugins:
                    if plugin == prj_plugin._name:
                        valid_plugins.append(prj_plugin)
        else:
            raise Exception(f'Provided data type for parameter plugins is not supported. Expects list, given:{type(plugins)}')

        # Check if any valid plugin was selected
        # TODO: does this need to be more selective? (What happens if we don't want any plugins?)
        if not valid_plugins:
            raise Exception(f'ERROR: Provided plugin/s:{plugins} were not registered in the project configuration!')

        # ToDo: currently, the generator API is under construction, which is why this only works for msdsl right now
        for plugin in valid_plugins:
            if plugin._name == 'msdsl':  # Pass generator inputs to plugin - note this is custom for each plugin
                if not self.filesets._functional_models:
                    # Use default functional models object for generation
                    func_model_default = FunctionalModel(files=os.path.join(self.args.input, 'gen.py'), config_path='',
                                                         name='main')
                    func_model_default.expand_paths()
                    func_model_default.set_gen_files_path(hdl_dir_root=self._prj_cfg.build_root_functional_models)
                    self.filesets._functional_models.append(func_model_default)

            plugin._set_generator_sources(generator_sources=self.filesets._functional_models)
            plugin.models()

    def build(self, *args, **kwargs):
        """
        Generate bitstream for FPGA target
        """

        shutil.rmtree(self._prj_cfg.build_root) # Remove target specific build dir to make sure there is no legacy
        mkdir_p(self._prj_cfg.build_root)
        self._setup_targets(target=self.act_fpga_target, gen_structures=True)

        # Check if active target is an FPGA target
        target = getattr(self, self.act_fpga_target)

        VivadoEmulation(target=target).build()

        # Build firmware for ctrl infrastructure if needed, first
        # waiting a short time for outputs from Vivado to sync
        # to disk.  This seems to be an issue with NFS.
        if target.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ:
            time.sleep(1)
            self._build_firmware(*args, **kwargs)

        statpro.statpro_update(statpro.FEATURES.anasymod_build_vivado)

    def emulate(self, server_addr=None, convert_waveform=True):
        """
        Program bitstream to FPGA and run simulation/emulation on FPGA

        :param server_addr: Address of Vivado hardware server used for communication to FPGA board
        """

        if server_addr is None:
            server_addr = self.args.server_addr

        # create target object, but don't generate instrumentation structure again in case target object does not exist yet
        if not hasattr(self, self.act_fpga_target):
            self._setup_targets(target=self.act_fpga_target)

        # check if bitstream was generated for active fpga target
        target = getattr(self, self.act_fpga_target)
        if not os.path.isfile(getattr(target, 'bitfile_path')):
            raise Exception(f'Bitstream for active FPGA target was not generated beforehand; please do so before running emulation.')

        # create sim result folders
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        if not os.path.exists(os.path.dirname(target.result_path_raw)):
            mkdir_p(os.path.dirname(target.result_path_raw))

        # run the emulation
        VivadoEmulation(target=target).run_FPGA(
            start_time=self.args.start_time,
            stop_time=self.args.stop_time,
            server_addr=server_addr
        )

        statpro.statpro_update(statpro.FEATURES.anasymod_emulate_vivado)

        # post-process results
        if convert_waveform:
            ConvertWaveform(
                result_path_raw=target.result_path_raw,
                result_type_raw=target.cfg.result_type_raw,
                result_path=target.cfg.vcd_path,
                str_cfg=target.str_cfg,
                float_type=self.float_type,
                dt_scale=self._prj_cfg.cfg.dt_scale
            )

    def launch(self, server_addr=None, debug=False):
        """
        Program bitstream to FPGA, setup control infrastructure and wait for interactive commands.

        :param server_addr: Address of Vivado hardware server used for communication to FPGA board
        :param debug: Enable or disable debug mode when running an interactive simulation
        """

        if server_addr is None:
            server_addr = self.args.server_addr

        # create target object, but don't generate instrumentation structure again in case target object does not exist yet
        if not hasattr(self, self.act_fpga_target):
            self._setup_targets(target=self.act_fpga_target, debug=debug)

        # check if bitstream was generated for active fpga target
        target = getattr(self, self.act_fpga_target)
        if not os.path.isfile(getattr(target, 'bitfile_path')):
            raise Exception(f'Bitstream for active FPGA target was not generated beforehand; please do so before running emulation.')

        # create sim result folders
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        if not os.path.exists(os.path.dirname(target.result_path_raw)):
            mkdir_p(os.path.dirname(target.result_path_raw))

        # launch the emulation
        ctrl_handle = VivadoEmulation(target=target).launch_FPGA(server_addr=server_addr)
        statpro.statpro_update(statpro.FEATURES.anasymod_emulate_vivado)

        # Return ctrl handle for interactive control
        return ctrl_handle

    def simulate(self, unit=None, id=None, convert_waveform=True, flags=None):
        """
        Run simulation on a pc target.  'flags' contains a list of simulator-specific
        flags, as a sort of escape hatch for features that are not yet supported
        natively through anasymod.
        """
        # Remove target-specific build dir to make sure there are no old files.
        # However, don't fail when certain files can't be removed, because that
        # might indicate that a waveform window is open
        shutil.rmtree(self._prj_cfg.build_root, ignore_errors=True)
        mkdir_p(self._prj_cfg.build_root)
        self._setup_targets(target=self.act_cpu_target, gen_structures=True)

        target = getattr(self, self.act_cpu_target)

        # create sim result folder
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        if not os.path.exists(os.path.dirname(target.result_path_raw)):
            mkdir_p(os.path.dirname(target.result_path_raw))

        # pick simulator
        sim_cls = {
            'icarus': IcarusSimulator,
            'vivado': VivadoSimulator,
            'xrun': XceliumSimulator
        }[self.args.simulator_name]

        # run simulation

        sim = sim_cls(target=target, flags=flags)

        if self.args.simulator_name == "xrun":
            sim.unit = unit
            sim.id = id

        sim.simulate()
        statpro.statpro_update(statpro.FEATURES.anasymod_sim + self.args.simulator_name)

        # post-process results
        if convert_waveform:
            ConvertWaveform(
                result_path_raw=target.result_path_raw,
                result_type_raw=target.cfg.result_type_raw,
                result_path=target.cfg.vcd_path,
                str_cfg=target.str_cfg,
                float_type=self.float_type,
                debug=self._prj_cfg.cfg.cpu_debug_mode,
                dt_scale=self._prj_cfg.cfg.dt_scale
            )

    def probe(self, name, emu_time=False):
        """
        Probe specified signal. Signal will be stored in a numpy array.
        """

        probeobj = self._setup_probeobj(target=getattr(self, self.args.active_target))
        return probeobj._probe(name=name, emu_time=emu_time)

    def probes(self):
        """
        Display all signals that were stored for specified target run (simulation or emulation)
        :return: list of signal names
        """

        probeobj = self._setup_probeobj(target=getattr(self, self.args.active_target))
        return probeobj._probes()

    def preserve(self, wave):
        """
        This function preserve the stepping of the waveform 'wave'. This is necessary, if limit checks should be
        conducted on the waveform later on.

        :param wave: 2d numpy.ndarray

        :return: 2d numpy.ndarray
        """
        temp_data = None
        wave_step =[]

        for d in wave.transpose():
            if temp_data is not None:
                if d[1] != temp_data:
                    wave_step.append([d[0],temp_data]) #old value with same timestep to preserve stepping
            wave_step.append(d)
            temp_data = d[1]

        try:
            #return np.array(wave_step, dtype='float').transpose()
            return np.array(wave_step, dtype='O').transpose()
        except:
            return np.array(wave_step, dtype='O').transpose()

    def view(self, result_file=None):
        """
        View results from selected target run.

        :param result_file: Path to the result file that shall be opened
        """

        root = self._prj_cfg.root

        if result_file is not None:
            result_file = expand_path(result_file, rel_path_reference=root)

        target = getattr(self, self.args.active_target)

        # pick viewer
        viewer_cls = {
            'gtkwave': GtkWaveViewer,
            'simvision': SimVisionViewer,
            'scansion': ScansionViewer
        }[self.args.viewer_name]

        # set config file location for GTKWave
        # TODO: clean this up; it's a bit messy...
        if isinstance(target, FPGATarget):
            gtkw_search_order = ['view_fpga.gtkw', 'view.gtkw']
        elif isinstance(target, CPUTarget):
            gtkw_search_order = ['view_sim.gtkw', 'view.gtkw']
        else:
            gtkw_search_order = ['view.gtkw']

        for basename in gtkw_search_order:
            candidate_path = os.path.join(root, basename)
            if os.path.isfile(candidate_path):
                self._prj_cfg.gtkwave_config.gtkw_config = candidate_path
                break
        else:
            self._prj_cfg.gtkwave_config.gtkw_config = None

        # set config file location for SimVision
        self._prj_cfg.simvision_config.svcf_config = os.path.join(root, 'view.svcf')

        # run viewer
        viewer = viewer_cls(target=target)
        viewer.view(result_file=result_file)

    def pack_results(self, target=None):
        """
        Pack target-specific build folder into a zip and store it in project root directory. All project related .yaml
        config files will also be stored in the bundle.

        :param target: Specify target, that shall be stored, by default the currently active target will be used.
        """

        # Specify target-specific paths
        if target is None:
            build_target_root = self._prj_cfg.build_root
            clks_file_path = getattr(getattr(getattr(self, self.args.active_target), 'str_cfg'), '_clks_file_path')
            simctrl_file_path = getattr(getattr(getattr(self, self.args.active_target), 'str_cfg'), '_simctrl_file_path')
        elif target in self.fpga_targets + self.cpu_targets:
            self._prj_cfg._update_build_root(active_target=target)
            build_target_root = self._prj_cfg.build_root
            clks_file_path = getattr(getattr(getattr(self, target), 'str_cfg'), '_clks_file_path')
            simctrl_file_path = getattr(getattr(getattr(self, target), 'str_cfg'), '_simctrl_file_path')
            self._prj_cfg._update_build_root(active_target=self.args.active_target)
        else:
            raise Exception(f'ERROR: Provided target:{target} does not exist in current project!')

        # Copy any .yaml config files from project to target-specific build root.
        config_files = [self.cfgfile_path, clks_file_path, simctrl_file_path]
        config_files.append(config for config in self.filesets._config_paths)

        for config_file in config_files:
            if os.path.isfile(config_file):
                dst = os.path.join(build_target_root, 'configs', os.path.relpath(config_file, self.args.input))
                try:
                    shutil.copyfile(config_file, dst)
                except:
                    raise Exception(f'ERROR: File:{config_file} could not be copied to:{dst}!')

        # Zip target-specific build folder and copy to project root
        shutil.make_archive(base_name=os.path.basename(self.args.input) + '_!_' + str(target if target is not None else self.args.active_target) + '_!_bundle',
                            base_dir=build_target_root,
                            format='zip',
                            root_dir=self.args.input)

    def unpack_results(self, bundle_path, force=False):
        """
        Unpack target-specific result bundle to build root, in order to view results or do further post-processing
        without having to run simulations again. In case force attribute is set to true, all .yaml config files stored
        in bundle will copied to project root and overwrite existing ones.

        :param bundle_path: Path to result bundle
        :param force: Set this attribute to True to make sure .yaml configs in bundle will replace currently existing ones.
        """

        # Unpack zip to target-specific root directory, deleting folder in case it already existed
        target = os.path.basename(bundle_path).split('_!_')[1]
        build_target_root = os.path.join(self.args.input, 'build', target)

        if os.path.exists(build_target_root):
            shutil.rmtree(build_target_root)

        with zipfile.ZipFile(bundle_path, "r") as zip_ref:
            zip_ref.extractall(path=build_target_root)

        # Compare if config .yaml files are different in bundle and current project
        bundle_config_files = []
        [bundle_config_files.append(config_file) for config_file in os.path.join(build_target_root, 'configs')]

        if target in self.fpga_targets + self.cpu_targets:
            self._prj_cfg._update_build_root(active_target=target)
            clks_file_path = getattr(getattr(getattr(self, str(target)), 'str_cfg'), '_clks_file_path')
            simctrl_file_path = getattr(getattr(getattr(self, str(target)), 'str_cfg'), '_simctrl_file_path')
            self._prj_cfg._update_build_root(active_target=self.args.active_target)
        else:
            raise Exception(f'ERROR: Provided target:{target} does not exist in current project!')

        orig_config_files = [self.cfgfile_path, clks_file_path, simctrl_file_path]
        orig_config_files.append(config for config in self.filesets._config_paths)

        for orig_config_file in orig_config_files:
            exists = False
            for bundle_config_file in bundle_config_files:
                if os.path.basename(orig_config_file) == os.path.basename(bundle_config_file):
                    exists = True
                    if not filecmp.cmp(orig_config_file, bundle_config_file):
                        print(f'WARNING: Config:{os.path.basename(orig_config_file)} does not match with config in bundle, project setup changed!')
                    if force:  # Depending on force flag overwrite config .yaml files in case they differ from the ones in current project
                        shutil.rmtree(orig_config_file)
                        try:
                            shutil.copyfile(bundle_config_file, orig_config_file)
                        except:
                            raise Exception(f'ERROR: File:{bundle_config_file} could not be copied to:{orig_config_file}!')
            if not exists:
                    print(f'WARNING: Config file:{os.path.basename(orig_config_file)} does not exist in bundle, project setup changed!')

##### Utility Functions

    def _parse_args(self):
        """
        Read command line arguments. This supports convenient usage from command shell e.g.:
        python analysis.py -i filter --models --sim --view

        -i, --input: Path to project root directory of the project that shall be opened and worked with.
            default=None

        --simulator_name: Simulator that shall be used for logic simulation.
            default=icarus for windows, xrun for linux

        --synthesizer_name: Synthesis engine that shall be used for FPGA synthesis.
            default=vivado

        --viewer_name: Waveform viewer that shall be used for viewing result waveforms.
            default=gtkwave for windows, simvision for linux

        --active_target: Target that shall be actively used.
            default='sim'

        --launch: Launch the FPGA simulation/emulation by programming the bitstream and preparing the control interface for interactive use.

        --sim: Execute logic simulation for selected simulation target.

        --view: Open results in selected waveform viewer.

        --build: Synthesize, run P&R and generate bitstream for selected target.

        --emulate: Execute FPGA run for selected target.

        --start_time: Start time for FPGA simulation.
            default=0

        --server_addr: Hardware server address for FPGA simulation. This is necessary for connecting to a vivado
            hardware server from linux, that was setup under windows.
            default=None

        --stop_time: Stop time for FPGA simulation
            default=None

        --preprocess_only: For icarus only, this will nur run the simulation, but only compile the netlist.

        --models: Generate functional models for selected project.

        """

        parser = ArgumentParser()

        # if the Cadence tools are available, use those as defaults instead
        try:
            x = XceliumConfig(None).xrun
            default_simulator_name = 'xrun' if x is not None else 'icarus'
        except:
            default_simulator_name = 'icarus'

        try:
            s = SimVisionConfig(None).simvision
            default_viewer_name = 'simvision' if s is not None else 'gtkwave'
        except:
            default_viewer_name = 'gtkwave'
            pass


        parser.add_argument('-i', '--input', type=str, default=None)
        parser.add_argument('--simulator_name', type=str, default=default_simulator_name)
        parser.add_argument('--synthesizer_name', type=str, default='vivado')
        parser.add_argument('--viewer_name', type=str, default=default_viewer_name)
        parser.add_argument('--active_target', type=str, default='sim')
        parser.add_argument('--unit', type=str, default=None)
        parser.add_argument('--id', type=str, default=None)
        parser.add_argument('--sim', action='store_true')
        parser.add_argument('--view', action='store_true')
        parser.add_argument('--build', action='store_true')
        parser.add_argument('--emulate', action='store_true')
        parser.add_argument('--launch', action='store_true')
        parser.add_argument('--start_time', type=float, default=0)
        parser.add_argument('--server_addr', type=str, default=None)
        parser.add_argument('--stop_time', type=float, default=None)
        parser.add_argument('--preprocess_only', action='store_true')
        parser.add_argument('--models', action='store_true')

        self.args, _ = parser.parse_known_args()

    def _setup_filesets(self):
        """
        Finalize filesets for the project. Before this function is called, all sources should have been added to the
        project, either via source.yaml files/plugin-specific includes, or interactively via the add_sources function.

        Note: Do not add more sources to the project after this function has been run; They will only be considered,
        if this function is executed again afterwards.
        """

        # Read source.yaml files and store in fileset object
        default_filesets = ['default'] + self.cpu_targets + self.fpga_targets
        self.filesets = Filesets(root=self.args.input,
                                 default_filesets=default_filesets,
                                 root_func_models=self._prj_cfg.build_root_functional_models)
        self.filesets.read_filesets()

        # Add Defines and Sources from plugins
        for plugin in self._plugins:
            plugin._setup_sources()
            plugin._setup_defines()
            self.filesets._defines += plugin._dump_defines()
            self.filesets._verilog_sources += plugin._dump_verilog_sources()
            self.filesets._verilog_headers += plugin._dump_verilog_headers()
            self.filesets._vhdl_sources += plugin._dump_vhdl_sources()

        # Add custom source and define objects here e.g.:
        config_path = os.path.join(self.args.input, 'source.yaml')

        # Add some default files depending on whether there is a custom top level
        for fileset in self.cpu_targets + self.fpga_targets:
            try:
                custom_top = self.cfg_file[ConfigSections.CPU_TARGET][fileset]['custom_top'] if fileset in self.cpu_targets else self.cfg_file[ConfigSections.FPGA_TARGET][fileset]['custom_top']
                print(f'Using custom top for fileset {fileset}.')
            except:
                custom_top = False

            if not custom_top:
                #ToDo: check if file inclusion should be target specific -> less for simulation only for example
                self.filesets.add_source(source=VerilogSource(files=os.path.join(self.args.input, 'tb.sv'),
                                                              config_path=config_path,
                                                              fileset=fileset,
                                                              name='tb'))

        # Set define variables specifying the emulator control architecture
        # TODO: find a better place for these operations, and try to avoid directly accessing the config dictionary
        for fileset in self.cpu_targets + self.fpga_targets:
            try:
                top_module = self.cfg_file[ConfigSections.CPU_TARGET][fileset]['top_module'] if fileset in self.cpu_targets else self.cfg_file[ConfigSections.FPGA_TARGET][fileset]['top_module']
            except:
                top_module = 'top'

            print(f'Using top module {top_module} for fileset {fileset}.')
            self.filesets.add_define(define=Define(name='CLK_MSDSL', value=f'{top_module}.clk_default_osc', fileset=fileset))
            self.filesets.add_define(define=Define(name='CKE_MSDSL', value=f'{top_module}.emu_cke', fileset=fileset))
            self.filesets.add_define(define=Define(name='RST_MSDSL', value=f'{top_module}.emu_rst', fileset=fileset))
            self.filesets.add_define(define=Define(name='DT_WIDTH', value=f'{self._prj_cfg.cfg.dt_width}', fileset=fileset))
            self.filesets.add_define(define=Define(name='DT_SCALE', value=f'{self._prj_cfg.cfg.dt_scale}', fileset=fileset))
            self.filesets.add_define(define=Define(name='TIME_WIDTH', value=f'{self._prj_cfg.cfg.time_width}', fileset=fileset))
            self.filesets.add_define(define=Define(name='EMU_DT', value=f'{self._prj_cfg.cfg.dt}', fileset=fileset))
            self.filesets.add_define(define=Define(name='DEC_WIDTH', value=f'{self._prj_cfg.cfg.dec_bits}', fileset=fileset))
            self.filesets.add_define(define=Define(name='EMU_CLK_FREQ', value=f'{self._prj_cfg.cfg.emu_clk_freq}', fileset=fileset))

        # Set cpu target specific defines
        for fileset in self.cpu_targets:
            self.filesets.add_define(define=Define(name='SIMULATION_MODE_MSDSL', fileset=fileset))

    def _setup_targets(self, target, gen_structures=False, debug=False):
        """
        Setup targets for project.
        This may differ from one project to another and needs customization.
        1. Create target object for each target that is supported in project
        2. Assign filesets to all target objects of the project
        """

        if not self.project_sources_finalized:
            # Populate the fileset dict which will be used to copy data to target object and store in filesets variable
            self.filesets.populate_fileset_dict()
            self.project_sources_finalized = True

        filesets = self.filesets.fileset_dict

        if target in self.cpu_targets:
            #######################################################
            # Create and setup simulation target
            #######################################################
            self.__setattr__(target, CPUTarget(prj_cfg=self._prj_cfg, plugins=self._plugins, name=target, float_type=self.float_type))
            getattr(getattr(self, target), 'assign_fileset')(fileset=filesets['default'])
            if target in filesets:
                getattr(getattr(self, target), 'assign_fileset')(fileset=filesets[target])

            # Update simulation target specific configuration
            getattr(getattr(getattr(self, target), 'cfg'), 'update_config')(subsection=target)
            getattr(getattr(self, target), 'set_tstop')()
            getattr(getattr(self, target), 'update_structure_config')()
            if (not getattr(getattr(getattr(self, target), 'cfg'), 'custom_top') and gen_structures):
                getattr(getattr(self, target), 'gen_structure')()

        elif target in self.fpga_targets:
            #######################################################
            # Create and setup FPGA target
            #######################################################
            self.__setattr__(target, FPGATarget(prj_cfg=self._prj_cfg, plugins=self._plugins, name=target, float_type=self.float_type))
            getattr(getattr(self, target), 'assign_fileset')(fileset=filesets['default'])
            if target in filesets:
                getattr(getattr(self, target), 'assign_fileset')(fileset=filesets[target])

            # Update fpga target specific configuration
            getattr(getattr(getattr(self, target), 'cfg'), 'update_config')(subsection=target)
            getattr(getattr(self, target), 'set_tstop')()
            getattr(getattr(self, target), 'update_structure_config')()
            if not getattr(getattr(getattr(self, target), 'cfg'), 'custom_top'):
                getattr(getattr(self, target), 'setup_ctrl_ifc')(debug=debug)
                if gen_structures:
                    getattr(getattr(self, target), 'gen_structure')()

        # Copy generated sources by plugin from plugin build_root to target-specific build_root
        for plugin in self._plugins:
            try:
                dst = os.path.join(self._prj_cfg.build_root, os.path.relpath(plugin._build_root, self._prj_cfg.build_root_base))
                if os.path.exists(dst):
                    shutil.rmtree(dst)
                shutil.copytree(plugin._build_root, dst)
            except OSError as exc:
                if exc.errno == errno.ENOTDIR:
                    shutil.copy(plugin._build_root, self._prj_cfg.build_root)
                else:
                    raise Exception(f'ERROR: Could not copy from root:{plugin._build_root} to {self._prj_cfg.build_root}')

        # Indication that project setup for active target is complete
        self._setup_finished[target] = True

    def _setup_probeobj(self, target: Union[FPGATarget, CPUTarget]):
        """
        Check if the requested probe obj in the target object already exists, in not create one.
        Return the probe object.

        :param target: Target that signals shall be extracted from
        :return: probe object that was selected in target object
        """

        # specify probe obj name, specific to selected simulator/synthesizer
        if isinstance(target, FPGATarget):
            target_name = f"prb_{self.args.synthesizer_name}"
        elif isinstance(target, CPUTarget):
            target_name = f"prb_{self.args.simulator_name}"
        else:
            raise ValueError(f"Provided target type:{target} is not supported")

        # check if probe obj is already existing, if not, instantiate one

        #ToDo: In future it should be also possible to instantiate different probe objects, depending on data format that shall be read in
        if target_name not in target.probes.keys():
            from anasymod.probe import ProbeVCD
            target.probes[target_name] = ProbeVCD(target=target)

        return target.probes[target_name]

    def _build_firmware(self, *args, **kwargs):
        # create target object, but don't generate instrumentation structure again in case target object does not exist yet
        if not hasattr(self, self.act_fpga_target):
            self._setup_targets(target=self.act_fpga_target, gen_structures=True)

        # check if bitstream was generated for active fpga target
        target = getattr(self, self.act_fpga_target)
        if not os.path.isfile(getattr(target, 'bitfile_path')):
            raise Exception(f'Bitstream for active FPGA target was not generated beforehand; please do so before running emulation.')

        # build the firmware
        XSCTEmulation(pcfg=target.prj_cfg,
                      content=target.content,
                      project_root=target.project_root,
                      top_module=target.cfg.top_module
                      ).build(*args, **kwargs)

    def _program_firmware(self, server_addr=None, *args, **kwargs):
        # create target object, but don't generate instrumentation structure again in case target object does not exist yet
        if not hasattr(self, self.act_fpga_target):
            self._setup_targets(target=self.act_fpga_target)

        # check if bitstream was generated for active fpga target
        target = getattr(self, self.act_fpga_target)
        if not os.path.isfile(getattr(target, 'bitfile_path')):
            raise Exception(f'Bitstream for active FPGA target was not generated beforehand; please do so before running emulation.')

        # program the firmware
        XSCTEmulation(pcfg=target.prj_cfg,
                      content=target.content,
                      project_root=target.project_root,
                      top_module=target.cfg.top_module
                      ).program(server_addr=server_addr, **kwargs)

def main():
    Analysis(op_mode='commandline')

if __name__ == '__main__':
    main()



class RunTimeEnvs:
    """
    Container including enums for supported Run Time Environments.
    """

    nocplusplus = "nocplusplus"
    pmmlegacy = "pmmlegacy"
    stdsc_gcc_32 = "stdsc_gcc_32"
    stdsc_gcc_64 = "stdsc_gcc_64"
    stdsc_vc14_32 = "stdsc_vc14_32"
    stdsc_vc14_64 = "stdsc_vc14_64"

class TestClassification:
    """
    Container including enums for supported Run Time Environments.
    """

    basic = "basic"
    weekend = "weekend"


class Target:
    """
    Container including enums for supported Targets.
    """

    sim_icarus = "sim_icarus"
    sim_vivado = "sim_vivado"
    build_vivado = "build_vivado"
    emulate_vivado = "emulate_vivado"
    sim_xcelium = "sim_xcelium"

